#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import json
import multiprocessing
import os
import time
import unittest
from typing import List, Optional
from unittest import mock
from unittest.mock import patch, MagicMock

from evdev.ecodes import (
    EV_KEY,
    EV_ABS,
    ABS_HAT0X,
    KEY_COMMA,
    BTN_TOOL_DOUBLETAP,
    KEY_A,
    REL_WHEEL,
    REL_X,
    ABS_X,
    REL_HWHEEL,
    BTN_LEFT,
)

from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.groups import _Groups, DeviceType
from inputremapper.gui.messages.message_broker import (
    MessageBroker,
    Signal,
)
from inputremapper.gui.messages.message_data import CombinationRecorded
from inputremapper.gui.messages.message_types import MessageType
from inputremapper.gui.reader_client import ReaderClient
from inputremapper.gui.reader_service import ReaderService, ContextDummy
from inputremapper.injection.global_uinputs import GlobalUInputs, UInput, FrontendUInput
from inputremapper.input_event import InputEvent
from tests.lib.constants import (
    EVENT_READ_TIMEOUT,
    START_READING_DELAY,
    MAX_ABS,
    MIN_ABS,
)
from tests.lib.fixtures import fixtures
from tests.lib.fixtures import new_event
from tests.lib.pipes import push_event, push_events
from tests.lib.spy import spy
from tests.lib.test_setup import test_setup

CODE_1 = 100
CODE_2 = 101
CODE_3 = 102


class Listener:
    def __init__(self):
        self.calls: List = []

    def __call__(self, data):
        self.calls.append(data)


def wait(func, timeout=1.0):
    """Wait for func to return True."""
    iterations = 0
    sleepytime = 0.1
    while not func():
        time.sleep(sleepytime)
        iterations += 1
        if iterations * sleepytime > timeout:
            break


@test_setup
class TestReaderAsyncio(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.reader_service = None
        self.groups = _Groups()
        self.message_broker = MessageBroker()
        self.reader_client = ReaderClient(self.message_broker, self.groups)

    def tearDown(self):
        try:
            self.reader_client.terminate()
        except (BrokenPipeError, OSError):
            pass

    async def create_reader_service(self, groups: Optional[_Groups] = None):
        # this will cause pending events to be copied over to the reader-service
        # process
        if not groups:
            groups = self.groups

        global_uinputs = GlobalUInputs(UInput)
        assert groups is not None
        self.reader_service = ReaderService(groups, global_uinputs)
        asyncio.ensure_future(self.reader_service.run())

    async def test_should_forward_to_dummy(self):
        # It forwards to a ForwardDummy, because the gui process
        # 1. can't inject and
        # 2. is not even supposed to inject anything
        # thanks to not using multiprocessing as opposed to the other tests, we can
        # access this stuff
        context = None
        original_create_event_pipeline = ReaderService._create_event_pipeline

        def remember_context(*args, **kwargs):
            nonlocal context
            context = original_create_event_pipeline(*args, **kwargs)
            return context

        with mock.patch.object(
            ReaderService,
            "_create_event_pipeline",
            remember_context,
        ):
            await self.create_reader_service()

            listener = Listener()
            self.message_broker.subscribe(MessageType.combination_recorded, listener)

            self.reader_client.set_group(self.groups.find(key="Foo Device 2"))
            self.reader_client.start_recorder()

            await asyncio.sleep(0.1)
            self.assertIsInstance(context, ContextDummy)

            with spy(
                context.forward_dummy,
                "write",
            ) as write_spy:
                events = [InputEvent.rel(REL_X, -1)]
                push_events(fixtures.foo_device_2_mouse, events)
                await asyncio.sleep(0.1)
                self.reader_client._read()
                self.assertEqual(0, len(listener.calls))

                # we want `write` to be called on the forward_dummy, because we want
                # those events to just disappear.
                self.assertEqual(write_spy.call_count, len(events))
                self.assertEqual([call[0] for call in write_spy.call_args_list], events)


@test_setup
class TestReaderMultiprocessing(unittest.TestCase):
    def setUp(self):
        self.reader_service_process = None
        self.groups = _Groups()
        self.message_broker = MessageBroker()
        self.global_uinputs = GlobalUInputs(UInput)
        self.reader_client = ReaderClient(self.message_broker, self.groups)

    def tearDown(self):
        try:
            self.reader_client.terminate()
        except (BrokenPipeError, OSError):
            pass

        if self.reader_service_process is not None:
            self.reader_service_process.join(timeout=1)
            if self.reader_service_process.is_alive():
                self.reader_service_process.terminate()

    def create_reader_service(self, groups: Optional[_Groups] = None):
        # this will cause pending events to be copied over to the reader-service
        # process
        if not groups:
            groups = self.groups

        def start_reader_service():
            # this is a new process, so create a new event loop, and all dependencies
            # from scratch, or something
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            global_uinputs = GlobalUInputs(FrontendUInput)
            global_uinputs.reset()
            reader_service = ReaderService(groups, global_uinputs)
            loop.run_until_complete(reader_service.run())

        self.reader_service_process = multiprocessing.Process(
            target=start_reader_service
        )
        self.reader_service_process.start()
        time.sleep(0.1)

    def test_reading(self):
        l1 = Listener()
        l2 = Listener()
        self.message_broker.subscribe(MessageType.combination_recorded, l1)
        self.message_broker.subscribe(MessageType.recording_finished, l2)
        self.create_reader_service()
        self.reader_client.set_group(self.groups.find(key="Foo Device 2"))
        self.reader_client.start_recorder()

        push_events(fixtures.foo_device_2_gamepad, [InputEvent.abs(ABS_HAT0X, 1)])
        # we need to sleep because we have two different fixtures,
        # which will lead to race conditions
        time.sleep(0.1)

        # relative axis events should be released automagically after 0.3s
        push_events(fixtures.foo_device_2_mouse, [InputEvent.rel(REL_X, 5)])
        time.sleep(0.1)
        # read all pending events. Having a glib mainloop would be better,
        # as it would call read automatically periodically
        self.reader_client._read()
        self.assertEqual(
            [
                CombinationRecorded(
                    InputCombination(
                        [
                            InputConfig(
                                type=3,
                                code=16,
                                analog_threshold=1,
                                origin_hash=fixtures.foo_device_2_gamepad.get_device_hash(),
                            )
                        ]
                    )
                ),
                CombinationRecorded(
                    InputCombination(
                        [
                            InputConfig(
                                type=3,
                                code=16,
                                analog_threshold=1,
                                origin_hash=fixtures.foo_device_2_gamepad.get_device_hash(),
                            ),
                            InputConfig(
                                type=2,
                                code=0,
                                analog_threshold=1,
                                origin_hash=fixtures.foo_device_2_mouse.get_device_hash(),
                            ),
                        ]
                    )
                ),
            ],
            l1.calls,
        )

        # release the hat switch should emit the recording finished event
        # as both the hat and relative axis are released by now
        push_events(fixtures.foo_device_2_gamepad, [InputEvent.abs(ABS_HAT0X, 0)])
        time.sleep(0.3)
        self.reader_client._read()
        self.assertEqual([Signal(MessageType.recording_finished)], l2.calls)

    def test_should_release_relative_axis(self):
        # the timeout is set to 0.3s
        l1 = Listener()
        l2 = Listener()
        self.message_broker.subscribe(MessageType.combination_recorded, l1)
        self.message_broker.subscribe(MessageType.recording_finished, l2)
        self.create_reader_service()
        self.reader_client.set_group(self.groups.find(key="Foo Device 2"))
        self.reader_client.start_recorder()

        push_events(fixtures.foo_device_2_mouse, [InputEvent.rel(REL_X, -5)])
        time.sleep(0.1)
        self.reader_client._read()

        self.assertEqual(
            [
                CombinationRecorded(
                    InputCombination(
                        [
                            InputConfig(
                                type=2,
                                code=0,
                                analog_threshold=-1,
                                origin_hash=fixtures.foo_device_2_mouse.get_device_hash(),
                            )
                        ]
                    )
                )
            ],
            l1.calls,
        )
        self.assertEqual([], l2.calls)  # no stop recording yet

        time.sleep(0.3)
        self.reader_client._read()
        self.assertEqual([Signal(MessageType.recording_finished)], l2.calls)

    def test_should_not_trigger_at_low_speed_for_rel_axis(self):
        l1 = Listener()
        self.message_broker.subscribe(MessageType.combination_recorded, l1)
        self.create_reader_service()
        self.reader_client.set_group(self.groups.find(key="Foo Device 2"))
        self.reader_client.start_recorder()

        push_events(fixtures.foo_device_2_mouse, [InputEvent.rel(REL_X, -1)])
        time.sleep(0.1)
        self.reader_client._read()
        self.assertEqual(0, len(l1.calls))

    def test_should_trigger_wheel_at_low_speed(self):
        l1 = Listener()
        self.message_broker.subscribe(MessageType.combination_recorded, l1)
        self.create_reader_service()
        self.reader_client.set_group(self.groups.find(key="Foo Device 2"))
        self.reader_client.start_recorder()

        push_events(
            fixtures.foo_device_2_mouse,
            [InputEvent.rel(REL_WHEEL, -1), InputEvent.rel(REL_HWHEEL, 1)],
        )
        time.sleep(0.1)
        self.reader_client._read()

        self.assertEqual(
            [
                CombinationRecorded(
                    InputCombination(
                        [
                            InputConfig(
                                type=2,
                                code=8,
                                analog_threshold=-1,
                                origin_hash=fixtures.foo_device_2_mouse.get_device_hash(),
                            )
                        ]
                    )
                ),
                CombinationRecorded(
                    InputCombination(
                        [
                            InputConfig(
                                type=2,
                                code=8,
                                analog_threshold=-1,
                                origin_hash=fixtures.foo_device_2_mouse.get_device_hash(),
                            ),
                            InputConfig(
                                type=2,
                                code=6,
                                analog_threshold=1,
                                origin_hash=fixtures.foo_device_2_mouse.get_device_hash(),
                            ),
                        ]
                    )
                ),
            ],
            l1.calls,
        )

    def test_wont_emit_the_same_combination_twice(self):
        l1 = Listener()
        self.message_broker.subscribe(MessageType.combination_recorded, l1)
        self.create_reader_service()
        self.reader_client.set_group(self.groups.find(key="Foo Device 2"))
        self.reader_client.start_recorder()

        push_events(fixtures.foo_device_2_keyboard, [InputEvent.key(KEY_A, 1)])
        time.sleep(0.1)
        self.reader_client._read()
        # the duplicate event should be ignored
        push_events(fixtures.foo_device_2_keyboard, [InputEvent.key(KEY_A, 1)])
        time.sleep(0.1)
        self.reader_client._read()

        self.assertEqual(
            [
                CombinationRecorded(
                    InputCombination(
                        [
                            InputConfig(
                                type=1,
                                code=30,
                                analog_threshold=1,
                                origin_hash=fixtures.foo_device_2_keyboard.get_device_hash(),
                            )
                        ]
                    )
                )
            ],
            l1.calls,
        )

    def test_should_read_absolut_axis(self):
        l1 = Listener()
        l2 = Listener()
        self.message_broker.subscribe(MessageType.combination_recorded, l1)
        self.message_broker.subscribe(MessageType.recording_finished, l2)
        self.create_reader_service()
        self.reader_client.set_group(self.groups.find(key="Foo Device 2"))
        self.reader_client.start_recorder()

        # over 30% should trigger
        push_events(
            fixtures.foo_device_2_gamepad,
            [InputEvent.abs(ABS_X, int(MAX_ABS * 0.4))],
        )
        time.sleep(0.1)
        self.reader_client._read()
        self.assertEqual(
            [
                CombinationRecorded(
                    InputCombination(
                        [
                            InputConfig(
                                type=3,
                                code=0,
                                analog_threshold=1,
                                origin_hash=fixtures.foo_device_2_gamepad.get_device_hash(),
                            )
                        ]
                    )
                )
            ],
            l1.calls,
        )
        self.assertEqual([], l2.calls)  # no stop recording yet

        # less the 30% should release
        push_events(
            fixtures.foo_device_2_gamepad,
            [InputEvent.abs(ABS_X, int(MAX_ABS * 0.2))],
        )
        time.sleep(0.1)
        self.reader_client._read()
        self.assertEqual(
            [
                CombinationRecorded(
                    InputCombination(
                        [
                            InputConfig(
                                type=3,
                                code=0,
                                analog_threshold=1,
                                origin_hash=fixtures.foo_device_2_gamepad.get_device_hash(),
                            )
                        ]
                    )
                )
            ],
            l1.calls,
        )
        self.assertEqual([Signal(MessageType.recording_finished)], l2.calls)

    def test_should_change_direction(self):
        l1 = Listener()
        self.message_broker.subscribe(MessageType.combination_recorded, l1)
        self.create_reader_service()
        self.reader_client.set_group(self.groups.find(key="Foo Device 2"))
        self.reader_client.start_recorder()

        push_event(fixtures.foo_device_2_keyboard, InputEvent.key(KEY_A, 1))
        time.sleep(0.1)
        push_event(
            fixtures.foo_device_2_gamepad, InputEvent.abs(ABS_X, int(MAX_ABS * 0.4))
        )
        time.sleep(0.1)
        push_event(fixtures.foo_device_2_keyboard, InputEvent.key(KEY_COMMA, 1))
        time.sleep(0.1)
        push_events(
            fixtures.foo_device_2_gamepad,
            [
                InputEvent.abs(ABS_X, int(MAX_ABS * 0.1)),
                InputEvent.abs(ABS_X, int(MIN_ABS * 0.4)),
            ],
        )
        time.sleep(0.1)
        self.reader_client._read()
        self.assertEqual(
            [
                CombinationRecorded(
                    InputCombination(
                        [
                            InputConfig(
                                type=EV_KEY,
                                code=KEY_A,
                                origin_hash=fixtures.foo_device_2_keyboard.get_device_hash(),
                            )
                        ]
                    )
                ),
                CombinationRecorded(
                    InputCombination(
                        [
                            InputConfig(
                                type=EV_KEY,
                                code=KEY_A,
                                origin_hash=fixtures.foo_device_2_keyboard.get_device_hash(),
                            ),
                            InputConfig(
                                type=EV_ABS,
                                code=ABS_X,
                                analog_threshold=1,
                                origin_hash=fixtures.foo_device_2_gamepad.get_device_hash(),
                            ),
                        ]
                    )
                ),
                CombinationRecorded(
                    InputCombination(
                        [
                            InputConfig(
                                type=EV_KEY,
                                code=KEY_A,
                                origin_hash=fixtures.foo_device_2_keyboard.get_device_hash(),
                            ),
                            InputConfig(
                                type=EV_ABS,
                                code=ABS_X,
                                analog_threshold=1,
                                origin_hash=fixtures.foo_device_2_gamepad.get_device_hash(),
                            ),
                            InputConfig(
                                type=EV_KEY,
                                code=KEY_COMMA,
                                origin_hash=fixtures.foo_device_2_keyboard.get_device_hash(),
                            ),
                        ]
                    )
                ),
                CombinationRecorded(
                    InputCombination(
                        [
                            InputConfig(
                                type=EV_KEY,
                                code=KEY_A,
                                origin_hash=fixtures.foo_device_2_keyboard.get_device_hash(),
                            ),
                            InputConfig(
                                type=EV_ABS,
                                code=ABS_X,
                                analog_threshold=-1,
                                origin_hash=fixtures.foo_device_2_gamepad.get_device_hash(),
                            ),
                            InputConfig(
                                type=EV_KEY,
                                code=KEY_COMMA,
                                origin_hash=fixtures.foo_device_2_keyboard.get_device_hash(),
                            ),
                        ]
                    )
                ),
            ],
            l1.calls,
        )

    def test_change_device(self):
        l1 = Listener()
        self.message_broker.subscribe(MessageType.combination_recorded, l1)

        push_events(
            fixtures.foo_device_2_keyboard,
            [
                InputEvent.key(1, 1),
            ]
            * 10,
        )

        push_events(
            fixtures.bar_device,
            [
                InputEvent.key(2, 1),
                InputEvent.key(2, 0),
            ]
            * 3,
        )

        self.create_reader_service()
        self.reader_client.set_group(self.groups.find(key="Foo Device 2"))
        self.reader_client.start_recorder()
        time.sleep(0.1)
        self.reader_client._read()
        self.assertEqual(
            l1.calls[0].combination,
            InputCombination(
                [
                    InputConfig(
                        type=EV_KEY,
                        code=1,
                        origin_hash=fixtures.foo_device_2_keyboard.get_device_hash(),
                    )
                ]
            ),
        )

        self.reader_client.set_group(self.groups.find(name="Bar Device"))
        time.sleep(0.1)
        self.reader_client._read()

        # we did not get the event from the "Bar Device" because the group change
        # stopped the recording
        self.assertEqual(len(l1.calls), 1)

        self.reader_client.start_recorder()
        push_events(fixtures.bar_device, [InputEvent.key(2, 1)])
        time.sleep(0.1)
        self.reader_client._read()
        self.assertEqual(
            l1.calls[1].combination,
            InputCombination(
                [
                    InputConfig(
                        type=EV_KEY,
                        code=2,
                        origin_hash=fixtures.bar_device.get_device_hash(),
                    )
                ]
            ),
        )

    def test_reading_2(self):
        l1 = Listener()
        self.message_broker.subscribe(MessageType.combination_recorded, l1)
        # a combination of events
        push_events(
            fixtures.foo_device_2_keyboard,
            [
                new_event(EV_KEY, CODE_1, 1, 10000.1234),
                new_event(EV_KEY, CODE_3, 1, 10001.1234),
            ],
        )

        pipe = multiprocessing.Pipe()

        def refresh():
            # from within the reader-service process notify this test that
            # refresh was called as expected
            pipe[1].send("refreshed")

        groups = _Groups()
        groups.refresh = refresh
        self.create_reader_service(groups)

        self.reader_client.set_group(self.groups.find(key="Foo Device 2"))
        self.reader_client.start_recorder()

        # sending anything arbitrary does not stop the reader-service
        self.reader_client._commands_pipe.send(856794)
        time.sleep(0.2)
        push_events(
            fixtures.foo_device_2_gamepad,
            [new_event(EV_ABS, ABS_HAT0X, -1, 10002.1234)],
        )
        time.sleep(0.1)
        # but it makes it look for new devices because maybe its list of
        # self.groups is not up-to-date
        self.assertTrue(pipe[0].poll())
        self.assertEqual(pipe[0].recv(), "refreshed")

        self.reader_client._read()
        self.assertEqual(
            l1.calls[-1].combination,
            InputCombination(
                [
                    InputConfig(
                        type=EV_KEY,
                        code=CODE_1,
                        origin_hash=fixtures.foo_device_2_keyboard.get_device_hash(),
                    ),
                    InputConfig(
                        type=EV_KEY,
                        code=CODE_3,
                        origin_hash=fixtures.foo_device_2_keyboard.get_device_hash(),
                    ),
                    InputConfig(
                        type=EV_ABS,
                        code=ABS_HAT0X,
                        analog_threshold=-1,
                        origin_hash=fixtures.foo_device_2_gamepad.get_device_hash(),
                    ),
                ]
            ),
        )

    def test_blacklisted_events(self):
        l1 = Listener()
        self.message_broker.subscribe(MessageType.combination_recorded, l1)

        push_events(
            fixtures.foo_device_2_mouse,
            [
                InputEvent.key(BTN_TOOL_DOUBLETAP, 1),
                InputEvent.key(BTN_LEFT, 1),
                InputEvent.key(BTN_TOOL_DOUBLETAP, 1),
            ],
            force=True,
        )
        self.create_reader_service()
        self.reader_client.set_group(self.groups.find(key="Foo Device 2"))
        self.reader_client.start_recorder()
        time.sleep(0.1)
        self.reader_client._read()
        self.assertEqual(
            l1.calls[-1].combination,
            InputCombination(
                [
                    InputConfig(
                        type=EV_KEY,
                        code=BTN_LEFT,
                        origin_hash=fixtures.foo_device_2_mouse.get_device_hash(),
                    )
                ]
            ),
        )

    def test_ignore_value_2(self):
        l1 = Listener()
        self.message_broker.subscribe(MessageType.combination_recorded, l1)
        # this is not a combination, because (EV_KEY CODE_3, 2) is ignored
        push_events(
            fixtures.foo_device_2_gamepad,
            [InputEvent.abs(ABS_HAT0X, 1), InputEvent.key(CODE_3, 2)],
            force=True,
        )
        self.create_reader_service()
        self.reader_client.set_group(self.groups.find(key="Foo Device 2"))
        self.reader_client.start_recorder()
        time.sleep(0.2)
        self.reader_client._read()
        self.assertEqual(
            l1.calls[-1].combination,
            InputCombination(
                [
                    InputConfig(
                        type=EV_ABS,
                        code=ABS_HAT0X,
                        analog_threshold=1,
                        origin_hash=fixtures.foo_device_2_gamepad.get_device_hash(),
                    )
                ]
            ),
        )

    def test_reading_ignore_up(self):
        l1 = Listener()
        self.message_broker.subscribe(MessageType.combination_recorded, l1)
        push_events(
            fixtures.foo_device_2_keyboard,
            [
                new_event(EV_KEY, CODE_1, 0, 10),
                new_event(EV_KEY, CODE_2, 1, 11),
                new_event(EV_KEY, CODE_3, 0, 12),
            ],
        )
        self.create_reader_service()
        self.reader_client.set_group(self.groups.find(key="Foo Device 2"))
        self.reader_client.start_recorder()
        time.sleep(0.1)
        self.reader_client._read()
        self.assertEqual(
            l1.calls[-1].combination,
            InputCombination(
                [
                    InputConfig(
                        type=EV_KEY,
                        code=CODE_2,
                        origin_hash=fixtures.foo_device_2_keyboard.get_device_hash(),
                    )
                ]
            ),
        )

    def test_wrong_device(self):
        l1 = Listener()
        self.message_broker.subscribe(MessageType.combination_recorded, l1)

        push_events(
            fixtures.foo_device_2_keyboard,
            [
                InputEvent.key(CODE_1, 1),
                InputEvent.key(CODE_2, 1),
                InputEvent.key(CODE_3, 1),
            ],
        )
        self.create_reader_service()
        self.reader_client.set_group(self.groups.find(name="Bar Device"))
        self.reader_client.start_recorder()
        time.sleep(EVENT_READ_TIMEOUT * 5)
        self.reader_client._read()
        self.assertEqual(len(l1.calls), 0)

    def test_inputremapper_devices(self):
        # Don't read from inputremapper devices, their keycodes are not
        # representative for the original key. As long as this is not
        # intentionally programmed it won't even do that. But it was at some
        # point.
        l1 = Listener()
        self.message_broker.subscribe(MessageType.combination_recorded, l1)
        push_events(
            fixtures.input_remapper_bar_device,
            [
                InputEvent.key(CODE_1, 1),
                InputEvent.key(CODE_2, 1),
                InputEvent.key(CODE_3, 1),
            ],
        )
        self.create_reader_service()
        self.reader_client.set_group(self.groups.find(name="Bar Device"))
        self.reader_client.start_recorder()
        time.sleep(EVENT_READ_TIMEOUT * 5)
        self.reader_client._read()
        self.assertEqual(len(l1.calls), 0)

    def test_terminate(self):
        self.create_reader_service()
        self.reader_client.set_group(self.groups.find(key="Foo Device 2"))

        push_events(fixtures.foo_device_2_keyboard, [InputEvent.key(CODE_3, 1)])
        time.sleep(START_READING_DELAY + EVENT_READ_TIMEOUT)
        self.assertTrue(self.reader_client._results_pipe.poll())

        self.reader_client.terminate()
        time.sleep(EVENT_READ_TIMEOUT)
        self.assertFalse(self.reader_client._results_pipe.poll())

        # no new events arrive after terminating
        push_events(fixtures.foo_device_2_keyboard, [InputEvent.key(CODE_3, 1)])
        time.sleep(EVENT_READ_TIMEOUT * 3)
        self.assertFalse(self.reader_client._results_pipe.poll())

    def test_are_new_groups_available(self):
        l1 = Listener()
        self.message_broker.subscribe(MessageType.groups, l1)
        self.create_reader_service()
        self.reader_client.groups.set_groups([])

        time.sleep(0.1)  # let the reader-service send the groups
        # read stuff from the reader-service, which includes the devices
        self.assertEqual("[]", self.reader_client.groups.dumps())
        self.reader_client._read()

        self.assertEqual(
            self.reader_client.groups.dumps(),
            json.dumps(
                [
                    json.dumps(
                        {
                            "paths": [
                                "/dev/input/event1",
                            ],
                            "names": ["Foo Device"],
                            "types": [DeviceType.KEYBOARD],
                            "key": "Foo Device",
                        }
                    ),
                    json.dumps(
                        {
                            "paths": [
                                "/dev/input/event11",
                                "/dev/input/event10",
                                "/dev/input/event13",
                                "/dev/input/event15",
                            ],
                            "names": [
                                "Foo Device foo",
                                "Foo Device",
                                "Foo Device",
                                "Foo Device bar",
                            ],
                            "types": [
                                DeviceType.GAMEPAD,
                                DeviceType.KEYBOARD,
                                DeviceType.MOUSE,
                            ],
                            "key": "Foo Device 2",
                        }
                    ),
                    json.dumps(
                        {
                            "paths": ["/dev/input/event20"],
                            "names": ["Bar Device"],
                            "types": [DeviceType.KEYBOARD],
                            "key": "Bar Device",
                        }
                    ),
                    json.dumps(
                        {
                            "paths": ["/dev/input/event30"],
                            "names": ["gamepad"],
                            "types": [DeviceType.GAMEPAD],
                            "key": "gamepad",
                        }
                    ),
                    json.dumps(
                        {
                            "paths": ["/dev/input/event40"],
                            "names": ["input-remapper Bar Device"],
                            "types": [DeviceType.KEYBOARD],
                            "key": "input-remapper Bar Device",
                        }
                    ),
                    json.dumps(
                        {
                            "paths": ["/dev/input/event52"],
                            "names": ["Qux/[Device]?"],
                            "types": [DeviceType.KEYBOARD],
                            "key": "Qux/[Device]?",
                        }
                    ),
                ]
            ),
        )

        self.assertEqual(len(l1.calls), 1)  # ensure we got the event

    def test_starts_the_service(self):
        # if ReaderClient can't see the ReaderService, a new ReaderService should
        # be started via pkexec
        with patch.object(ReaderService, "is_running", lambda: False):
            os_system_mock = MagicMock(return_value=0)
            with patch.object(os, "system", os_system_mock):
                # the status message enables the reader-client to see, that the
                # reader-service has started
                self.reader_client._results_pipe.send(
                    {"type": "status", "message": "ready"}
                )
                self.reader_client._send_command("foo")
                os_system_mock.assert_called_once_with(
                    "pkexec input-remapper-control --command start-reader-service -d"
                )

    def test_wont_start_the_service(self):
        # already running, no call to os.system
        with patch.object(ReaderService, "is_running", lambda: True):
            mock = MagicMock(return_value=0)
            with patch.object(os, "system", mock):
                self.reader_client._send_command("foo")
                mock.assert_not_called()

    def test_reader_service_wont_start(self):
        # test for the "The reader-service did not start" message

        expected_msg = "The reader-service did not start"
        subscribe_mock = MagicMock()
        self.message_broker.subscribe(MessageType.status_msg, subscribe_mock)

        with patch.object(ReaderClient, "_timeout", 1):
            with patch.object(ReaderService, "is_running", lambda: False):
                os_system_mock = MagicMock(return_value=0)
                with patch.object(os, "system", os_system_mock):
                    self.reader_client._send_command("foo")
                    # no message is sent into _results_pipe, so the reader-client will
                    # think the reader-service didn't manage to start
                    os_system_mock.assert_called_once_with(
                        "pkexec input-remapper-control "
                        "--command start-reader-service -d"
                    )

        subscribe_mock.assert_called_once()
        status = subscribe_mock.call_args[0][0]
        self.assertEqual(status.msg, expected_msg)

    def test_reader_service_times_out(self):
        # after some time the reader-service just stops, to avoid leaving a hole
        # that exposes user-input forever
        with patch.object(ReaderService, "_maximum_lifetime", 1):
            self.create_reader_service()
            self.assertTrue(self.reader_service_process.is_alive())
            time.sleep(0.5)
            self.assertTrue(self.reader_service_process.is_alive())
            time.sleep(1)
            self.assertFalse(self.reader_service_process.is_alive())

    def test_reader_service_waits_for_client_to_finish(self):
        # if the client is currently reading, it waits a bit longer until the
        # client finishes reading
        with patch.object(ReaderService, "_maximum_lifetime", 1):
            self.create_reader_service()
            self.assertTrue(self.reader_service_process.is_alive())

            self.reader_client.set_group(self.groups.find(key="Foo Device 2"))
            self.reader_client.start_recorder()

            time.sleep(2)
            # still alive, without start_recorder it should have already exited
            self.assertTrue(self.reader_service_process.is_alive())

            self.reader_client.stop_recorder()

            time.sleep(1)
            self.assertFalse(self.reader_service_process.is_alive())

    def test_reader_service_wont_wait_forever(self):
        # if the client is reading forever, stop it after another timeout
        with patch.object(ReaderService, "_maximum_lifetime", 1):
            with patch.object(ReaderService, "_timeout_tolerance", 1):
                self.create_reader_service()
                self.assertTrue(self.reader_service_process.is_alive())

                self.reader_client.set_group(self.groups.find(key="Foo Device 2"))
                self.reader_client.start_recorder()

                time.sleep(1.5)
                # still alive, without start_recorder it should have already exited
                self.assertTrue(self.reader_service_process.is_alive())

                time.sleep(1)
                # now it stopped, even though the reader is still reading
                self.assertFalse(self.reader_service_process.is_alive())

    def test_reader_service_stops_with_broken_pipes(self):
        with patch.object(ReaderService, "pipes_exist", side_effect=lambda: False):
            self.create_reader_service()
            self.assertTrue(self.reader_service_process.is_alive())

            time.sleep(0.5)
            # still alive
            self.assertTrue(self.reader_service_process.is_alive())

            time.sleep(1)
            # now it stopped
            self.assertFalse(self.reader_service_process.is_alive())

        # It will not stop if pipes are ok
        with patch.object(ReaderService, "pipes_exist", side_effect=lambda: True):
            self.create_reader_service()
            self.assertTrue(self.reader_service_process.is_alive())

            time.sleep(1.5)
            # still alive
            self.assertTrue(self.reader_service_process.is_alive())


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


import unittest
from unittest.mock import patch

from evdev.ecodes import (
    EV_KEY,
    KEY_1,
    KEY_2,
    LED_CAPSL,
    LED_NUML,
)

from inputremapper.injection.macros.parse import Parser
from tests.lib.test_setup import test_setup
from tests.unit.test_macros.macro_test_base import DummyMapping, MacroTestBase


@test_setup
class TestLeds(MacroTestBase):
    async def test_if_capslock(self):
        macro = Parser.parse(
            "if_capslock(key(KEY_1), key(KEY_2))",
            self.context,
            DummyMapping,
            True,
        )
        self.assertEqual(self.count_child_macros(macro), 2)

        with patch.object(self.source_device, "leds", side_effect=lambda: [LED_NUML]):
            await macro.run(self.handler)
            self.assertListEqual(self.result, [(EV_KEY, KEY_2, 1), (EV_KEY, KEY_2, 0)])

        with patch.object(self.source_device, "leds", side_effect=lambda: [LED_CAPSL]):
            self.result = []
            await macro.run(self.handler)
            self.assertListEqual(self.result, [(EV_KEY, KEY_1, 1), (EV_KEY, KEY_1, 0)])

    async def test_if_numlock(self):
        macro = Parser.parse(
            "if_numlock(key(KEY_1), key(KEY_2))",
            self.context,
            DummyMapping,
            True,
        )
        self.assertEqual(self.count_child_macros(macro), 2)

        with patch.object(self.source_device, "leds", side_effect=lambda: [LED_NUML]):
            await macro.run(self.handler)
            self.assertListEqual(self.result, [(EV_KEY, KEY_1, 1), (EV_KEY, KEY_1, 0)])

        with patch.object(self.source_device, "leds", side_effect=lambda: [LED_CAPSL]):
            self.result = []
            await macro.run(self.handler)
            self.assertListEqual(self.result, [(EV_KEY, KEY_2, 1), (EV_KEY, KEY_2, 0)])

    async def test_if_numlock_no_else(self):
        macro = Parser.parse(
            "if_numlock(key(KEY_1))",
            self.context,
            DummyMapping,
            True,
        )
        self.assertEqual(self.count_child_macros(macro), 1)

        with patch.object(self.source_device, "leds", side_effect=lambda: [LED_CAPSL]):
            await macro.run(self.handler)
            self.assertListEqual(self.result, [])

        with patch.object(self.source_device, "leds", side_effect=lambda: [LED_NUML]):
            self.result = []
            await macro.run(self.handler)
            self.assertListEqual(self.result, [(EV_KEY, KEY_1, 1), (EV_KEY, KEY_1, 0)])

    async def test_if_capslock_no_then(self):
        macro = Parser.parse(
            "if_capslock(None, key(KEY_1))",
            self.context,
            DummyMapping,
            True,
        )
        self.assertEqual(self.count_child_macros(macro), 1)

        with patch.object(self.source_device, "leds", side_effect=lambda: [LED_CAPSL]):
            await macro.run(self.handler)
            self.assertListEqual(self.result, [])

        with patch.object(self.source_device, "leds", side_effect=lambda: [LED_NUML]):
            self.result = []
            await macro.run(self.handler)
            self.assertListEqual(self.result, [(EV_KEY, KEY_1, 1), (EV_KEY, KEY_1, 0)])

    async def test_raises_error(self):
        Parser.parse("if_capslock(else=key(KEY_A))", self.context)  # no error
        Parser.parse("if_capslock(key(KEY_A), None)", self.context)  # no error
        Parser.parse("if_capslock(key(KEY_A))", self.context)  # no error
        Parser.parse("if_capslock(then=key(KEY_A))", self.context)  # no error
        Parser.parse("if_numlock(else=key(KEY_A))", self.context)  # no error
        Parser.parse("if_numlock(key(KEY_A), None)", self.context)  # no error
        Parser.parse("if_numlock(key(KEY_A))", self.context)  # no error
        Parser.parse("if_numlock(then=key(KEY_A))", self.context)  # no error


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


"""Components used in multiple places."""


from __future__ import annotations


from gi.repository import Gtk

from typing import Optional

from inputremapper.configs.mapping import MappingData

from inputremapper.gui.controller import Controller
from inputremapper.gui.messages.message_broker import (
    MessageBroker,
    MessageType,
)
from inputremapper.gui.messages.message_data import GroupData, PresetData
from inputremapper.gui.utils import HandlerDisabled


class FlowBoxEntry(Gtk.ToggleButton):
    """A device that can be selected in the GUI.

    For example a keyboard or a mouse.
    """

    __gtype_name__ = "FlowBoxEntry"

    def __init__(
        self,
        message_broker: MessageBroker,
        controller: Controller,
        name: str,
        icon_name: Optional[str] = None,
    ):
        super().__init__()
        self.icon_name = icon_name
        self.message_broker = message_broker
        self._controller = controller

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        if icon_name:
            icon = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.DIALOG)
            box.add(icon)

        label = Gtk.Label()
        label.set_label(name)
        self.name = name

        # wrap very long names properly
        label.set_line_wrap(True)
        label.set_line_wrap_mode(2)
        # this affeects how many device entries fit next to each other
        label.set_width_chars(28)
        label.set_max_width_chars(28)

        box.add(label)

        box.set_margin_top(18)
        box.set_margin_bottom(18)
        box.set_homogeneous(True)
        box.set_spacing(12)

        # self.set_relief(Gtk.ReliefStyle.NONE)

        self.add(box)

        self.show_all()

        self.connect("toggled", self._on_gtk_toggle)

    def _on_gtk_toggle(self):
        raise NotImplementedError

    def show_active(self, active):
        """Show the active state without triggering anything."""
        with HandlerDisabled(self, self._on_gtk_toggle):
            self.set_active(active)


class FlowBoxWrapper:
    """A wrapper for a flowbox that contains FlowBoxEntry widgets."""

    def __init__(self, flowbox: Gtk.FlowBox):
        self._gui = flowbox

    def show_active_entry(self, name: Optional[str]):
        """Activate the togglebutton that matches the name."""
        for child in self._gui.get_children():
            flow_box_entry: FlowBoxEntry = child.get_children()[0]
            flow_box_entry.show_active(flow_box_entry.name == name)


class Breadcrumbs:
    """Writes a breadcrumbs string into a given label."""

    def __init__(
        self,
        message_broker: MessageBroker,
        label: Gtk.Label,
        show_device_group: bool = False,
        show_preset: bool = False,
        show_mapping: bool = False,
    ):
        self._message_broker = message_broker
        self._gui = label
        self._connect_message_listener()

        self.show_device_group = show_device_group
        self.show_preset = show_preset
        self.show_mapping = show_mapping

        self._group_key: str = ""
        self._preset_name: str = ""
        self._mapping_name: str = ""

        label.set_max_width_chars(50)
        label.set_line_wrap(True)
        label.set_line_wrap_mode(2)

        self._render()

    def _connect_message_listener(self):
        self._message_broker.subscribe(MessageType.group, self._on_group_changed)
        self._message_broker.subscribe(MessageType.preset, self._on_preset_changed)
        self._message_broker.subscribe(MessageType.mapping, self._on_mapping_changed)

    def _on_preset_changed(self, data: PresetData):
        self._preset_name = data.name or ""
        self._render()

    def _on_group_changed(self, data: GroupData):
        self._group_key = data.group_key
        self._render()

    def _on_mapping_changed(self, mapping_data: MappingData):
        self._mapping_name = mapping_data.format_name()
        self._render()

    def _render(self):
        label = []

        if self.show_device_group:
            label.append(self._group_key or "?")

        if self.show_preset:
            label.append(self._preset_name or "?")

        if self.show_mapping:
            label.append(self._mapping_name or "?")

        self._gui.set_label("  /  ".join(label))



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

"""Contains and manages mappings."""

from __future__ import annotations

import json
import os
from typing import (
    Tuple,
    Dict,
    List,
    Optional,
    Iterator,
    Type,
    TypeVar,
    Generic,
    overload,
)

from evdev import ecodes

from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.configs.mapping import Mapping, UIMapping
from inputremapper.configs.paths import PathUtils
from inputremapper.logging.logger import logger

MappingModel = TypeVar("MappingModel", bound=UIMapping)


class Preset(Generic[MappingModel]):
    """Contains and manages mappings of a single preset."""

    # workaround for typing: https://github.com/python/mypy/issues/4236
    @overload
    def __init__(self: Preset[Mapping], path: Optional[os.PathLike] = None): ...

    @overload
    def __init__(
        self,
        path: Optional[os.PathLike] = None,
        mapping_factory: Type[MappingModel] = ...,
    ): ...

    def __init__(
        self,
        path: Optional[os.PathLike] = None,
        mapping_factory=Mapping,
    ) -> None:
        self._mappings: Dict[InputCombination, MappingModel] = {}
        # a copy of mappings for keeping track of changes
        self._saved_mappings: Dict[InputCombination, MappingModel] = {}
        self._path: Optional[os.PathLike] = path

        # the mapping class which is used by load()
        self._mapping_factory: Type[MappingModel] = mapping_factory

    def __iter__(self) -> Iterator[MappingModel]:
        """Iterate over Mapping objects."""
        return iter(self._mappings.copy().values())

    def __len__(self) -> int:
        return len(self._mappings)

    def __bool__(self):
        # otherwise __len__ will be used which results in False for a preset
        # without mappings
        return True

    def has_unsaved_changes(self) -> bool:
        """Check if there are unsaved changed."""
        return self._mappings != self._saved_mappings

    def remove(self, combination: InputCombination) -> None:
        """Remove a mapping from the preset by providing the InputCombination."""

        if not isinstance(combination, InputCombination):
            raise TypeError(
                f"combination must by of type InputCombination, got {type(combination)}"
            )

        for permutation in combination.get_permutations():
            if permutation in self._mappings.keys():
                combination = permutation
                break
        try:
            mapping = self._mappings.pop(combination)
            mapping.remove_combination_changed_callback()
        except KeyError:
            logger.debug(
                "unable to remove non-existing mapping with combination = %s",
                combination,
            )
            pass

    def add(self, mapping: MappingModel) -> None:
        """Add a mapping to the preset."""
        for permutation in mapping.input_combination.get_permutations():
            if permutation in self._mappings:
                raise KeyError(
                    "A mapping with this input_combination: "
                    f"{permutation} already exists",
                )

        mapping.set_combination_changed_callback(self._combination_changed_callback)
        self._mappings[mapping.input_combination] = mapping

    def empty(self) -> None:
        """Remove all mappings and custom configs without saving.
        note: self.has_unsaved_changes() will report True
        """
        for mapping in self._mappings.values():
            mapping.remove_combination_changed_callback()
        self._mappings = {}

    def clear(self) -> None:
        """Remove all mappings and also self.path."""
        self.empty()
        self._saved_mappings = {}
        self.path = None

    def load(self) -> None:
        """Load from the mapping from the disc, clears all existing mappings."""
        logger.info('Loading preset from "%s"', self.path)

        if not self.path or not os.path.exists(self.path):
            raise FileNotFoundError(f'Tried to load non-existing preset "{self.path}"')

        self._saved_mappings = self._get_mappings_from_disc()
        self.empty()
        for mapping in self._saved_mappings.values():
            # use the public add method to make sure
            # the _combination_changed_callback is attached
            self.add(mapping.copy())

    def _is_mapped_multiple_times(self, input_combination: InputCombination) -> bool:
        """Check if the event combination maps to multiple mappings."""
        all_input_combinations = {mapping.input_combination for mapping in self}
        permutations = set(input_combination.get_permutations())
        union = permutations & all_input_combinations
        # if there are more than one matches, then there is a duplicate
        return len(union) > 1

    def _has_valid_input_combination(self, mapping: UIMapping) -> bool:
        """Check if the mapping has a valid input event combination."""
        is_a_combination = isinstance(mapping.input_combination, InputCombination)
        is_empty = mapping.input_combination == InputCombination.empty_combination()
        return is_a_combination and not is_empty

    def save(self) -> None:
        """Dump as JSON to self.path."""

        if not self.path:
            logger.debug("unable to save preset without a path set Preset.path first")
            return

        PathUtils.touch(self.path)
        if not self.has_unsaved_changes():
            logger.debug("Not saving unchanged preset")
            return

        logger.info("Saving preset to %s", self.path)

        preset_list = []
        saved_mappings = {}
        for mapping in self:
            if not mapping.is_valid():
                if not self._has_valid_input_combination(mapping):
                    # we save invalid mappings except for those with an invalid
                    # input_combination
                    logger.debug("Skipping invalid mapping %s", mapping)
                    continue

                if self._is_mapped_multiple_times(mapping.input_combination):
                    # todo: is this ever executed? it should not be possible to
                    #  reach this
                    logger.debug(
                        "skipping mapping with duplicate event combination %s",
                        mapping,
                    )
                    continue

            mapping_dict = mapping.dict(exclude_defaults=True)
            mapping_dict["input_combination"] = mapping.input_combination.to_config()
            combination = mapping.input_combination
            preset_list.append(mapping_dict)

            saved_mappings[combination] = mapping.copy()
            saved_mappings[combination].remove_combination_changed_callback()

        with open(self.path, "w") as file:
            json.dump(preset_list, file, indent=4)
            file.write("\n")

        self._saved_mappings = saved_mappings

    def is_valid(self) -> bool:
        return False not in [mapping.is_valid() for mapping in self]

    def get_mapping(
        self, combination: Optional[InputCombination]
    ) -> Optional[MappingModel]:
        """Return the Mapping that is mapped to this InputCombination."""
        if not combination:
            return None

        if not isinstance(combination, InputCombination):
            raise TypeError(
                f"combination must by of type InputCombination, got {type(combination)}"
            )

        for permutation in combination.get_permutations():
            existing = self._mappings.get(permutation)
            if existing is not None:
                return existing
        return None

    def dangerously_mapped_btn_left(self) -> bool:
        """Return True if this mapping disables BTN_Left."""
        if (ecodes.EV_KEY, ecodes.BTN_LEFT) not in [
            m.input_combination[0].type_and_code for m in self
        ]:
            return False

        values: List[str | Tuple[int, int] | None] = []
        for mapping in self:
            if mapping.output_symbol is None:
                continue
            values.append(mapping.output_symbol.lower())
            values.append(mapping.get_output_type_code())

        return (
            "btn_left" not in values
            or InputConfig.btn_left().type_and_code not in values
        )

    def _combination_changed_callback(
        self, new: InputCombination, old: InputCombination
    ) -> None:
        for permutation in new.get_permutations():
            if permutation in self._mappings.keys() and permutation != old:
                raise KeyError("combination already exists in the preset")
        self._mappings[new] = self._mappings.pop(old)

    def _update_saved_mappings(self) -> None:
        if self.path is None:
            return

        if not os.path.exists(self.path):
            self._saved_mappings = {}
            return
        self._saved_mappings = self._get_mappings_from_disc()

    def _get_mappings_from_disc(self) -> Dict[InputCombination, MappingModel]:
        mappings: Dict[InputCombination, MappingModel] = {}
        if not self.path:
            logger.debug("unable to read preset without a path set Preset.path first")
            return mappings

        if os.stat(self.path).st_size == 0:
            logger.debug("got empty file")
            return mappings

        with open(self.path, "r") as file:
            try:
                preset_list = json.load(file)
            except json.JSONDecodeError:
                logger.error("unable to decode json file: %s", self.path)
                return mappings

        for mapping_dict in preset_list:
            if not isinstance(mapping_dict, dict):
                logger.error(
                    "Expected mapping to be a dict: %s %s",
                    type(mapping_dict),
                    mapping_dict,
                )
                continue

            try:
                mapping = self._mapping_factory(**mapping_dict)
            except Exception as error:
                logger.error(
                    "failed to Validate mapping for %s: %s",
                    mapping_dict.get("input_combination"),
                    error,
                )
                continue

            mappings[mapping.input_combination] = mapping
        return mappings

    @property
    def path(self) -> Optional[os.PathLike]:
        return self._path

    @path.setter
    def path(self, path: Optional[os.PathLike]):
        if path != self.path:
            self._path = path
            self._update_saved_mappings()

    @property
    def name(self) -> Optional[str]:
        """The name of the preset."""
        if self.path:
            return os.path.basename(self.path).split(".")[0]
        return None



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import enum
from collections import namedtuple
from typing import Optional, Callable, Tuple, TypeVar, Union, Any, Dict

from evdev.ecodes import (
    EV_KEY,
    EV_ABS,
    EV_REL,
    REL_WHEEL,
    REL_HWHEEL,
    REL_HWHEEL_HI_RES,
    REL_WHEEL_HI_RES,
)
from packaging import version

from inputremapper.logging.logger import logger

try:
    from pydantic.v1 import (
        BaseModel,
        PositiveInt,
        confloat,
        conint,
        root_validator,
        validator,
        ValidationError,
        PositiveFloat,
        VERSION,
        BaseConfig,
    )
except ImportError:
    from pydantic import (
        BaseModel,
        PositiveInt,
        confloat,
        conint,
        root_validator,
        validator,
        ValidationError,
        PositiveFloat,
        VERSION,
        BaseConfig,
    )

from inputremapper.configs.input_config import InputCombination
from inputremapper.configs.keyboard_layout import keyboard_layout, DISABLE_NAME
from inputremapper.configs.validation_errors import (
    OutputSymbolUnknownError,
    SymbolNotAvailableInTargetError,
    OnlyOneAnalogInputError,
    TriggerPointInRangeError,
    OutputSymbolVariantError,
    MacroButTypeOrCodeSetError,
    SymbolAndCodeMismatchError,
    WrongMappingTypeForKeyError,
    MissingOutputAxisError,
    MissingMacroOrKeyError,
)
from inputremapper.gui.gettext import _
from inputremapper.gui.messages.message_types import MessageType
from inputremapper.injection.global_uinputs import GlobalUInputs
from inputremapper.injection.macros.parse import Parser
from inputremapper.utils import get_evdev_constant_name

# TODO: remove pydantic VERSION check as soon as we no longer support
#  Ubuntu 20.04 and with it the ancient pydantic 1.2

needs_workaround = version.parse(str(VERSION)) < version.parse("1.7.1")


EMPTY_MAPPING_NAME: str = _("Empty Mapping")

# If `1` is the default speed for EV_REL, how much does this value needs to be scaled
# up to get reasonable speeds for various EV_REL events?
# Mouse injection rates vary wildly, and so do the values.
REL_XY_SCALING: float = 60
WHEEL_SCALING: float = 1
# WHEEL_HI_RES always generates events with 120 times higher values than WHEEL
# https://www.kernel.org/doc/html/latest/input/event-codes.html?highlight=wheel_hi_res#ev-rel
WHEEL_HI_RES_SCALING: float = 120
# Those values are assuming a rate of 60hz
DEFAULT_REL_RATE: float = 60


class KnownUinput(str, enum.Enum):
    """The default targets."""

    KEYBOARD = "keyboard"
    MOUSE = "mouse"
    GAMEPAD = "gamepad"
    KEYBOARD_MOUSE = "keyboard + mouse"


class MappingType(str, enum.Enum):
    """What kind of output the mapping produces."""

    KEY_MACRO = "key_macro"
    ANALOG = "analog"


CombinationChangedCallback = Optional[
    Callable[[InputCombination, InputCombination], None]
]
MappingModel = TypeVar("MappingModel", bound="UIMapping")


class Cfg(BaseConfig):
    validate_assignment = True
    use_enum_values = True
    underscore_attrs_are_private = True
    json_encoders = {InputCombination: lambda v: v.json_key()}


class ImmutableCfg(Cfg):
    allow_mutation = False


class UIMapping(BaseModel):
    """Holds all the data for mapping an input action to an output action.

    The Preset contains multiple UIMappings.

    This mapping does not validate the structure of the mapping or macros, only basic
    values. It is meant to be used in the GUI where invalid mappings are expected.
    """

    if needs_workaround:
        __slots__ = ("_combination_changed",)

    # Required attributes
    # The InputEvent or InputEvent combination which is mapped
    input_combination: InputCombination = InputCombination.empty_combination()
    # The UInput to which the mapped event will be sent
    target_uinput: Optional[Union[str, KnownUinput]] = None

    # Either `output_symbol` or `output_type` and `output_code` is required
    # Only set if output is "Key or Macro":
    output_symbol: Optional[str] = None  # The symbol or macro string if applicable
    # "Analog Axis" or if preset edited manually to inject a code instead of a symbol:
    output_type: Optional[int] = None  # The event type of the mapped event
    output_code: Optional[int] = None  # The event code of the mapped event

    name: Optional[str] = None
    mapping_type: Optional[MappingType] = None

    # if release events will be sent to the forwarded device as soon as a combination
    # triggers see also #229
    release_combination_keys: bool = True

    # macro settings
    macro_key_sleep_ms: conint(ge=0) = 0  # type: ignore

    # Optional attributes for mapping Axis to Axis
    # The deadzone of the input axis
    deadzone: confloat(ge=0, le=1) = 0.1  # type: ignore
    gain: float = 1.0  # The scale factor for the transformation
    # The expo factor for the transformation
    expo: confloat(ge=-1, le=1) = 0  # type: ignore

    # when mapping to relative axis
    # The frequency [Hz] at which EV_REL events get generated
    rel_rate: PositiveInt = 60

    # when mapping from a relative axis:
    # the relative value at which a EV_REL axis is considered at its maximum. Moving
    # a mouse at 2x the regular speed would be considered max by default.
    rel_to_abs_input_cutoff: PositiveInt = 2

    # the time until a relative axis is considered stationary if no new events arrive
    release_timeout: PositiveFloat = 0.05
    # don't release immediately when a relative axis drops below the speed threshold
    # instead wait until it dropped for loger than release_timeout below the threshold
    force_release_timeout: bool = False

    # callback which gets called if the input_combination is updated
    if not needs_workaround:
        _combination_changed: Optional[CombinationChangedCallback] = None

    # use type: ignore, looks like a mypy bug related to:
    # https://github.com/samuelcolvin/pydantic/issues/2949
    def __init__(self, **kwargs):  # type: ignore
        super().__init__(**kwargs)
        if needs_workaround:
            object.__setattr__(self, "_combination_changed", None)

    def __setattr__(self, key: str, value: Any):
        """Call the combination changed callback
        if we are about to update the input_combination
        """
        if key != "input_combination" or self._combination_changed is None:
            if key == "_combination_changed" and needs_workaround:
                object.__setattr__(self, "_combination_changed", value)
                return
            super().__setattr__(key, value)
            return

        # the new combination is not yet validated
        try:
            new_combi = InputCombination.validate(value)
        except (ValueError, TypeError) as exception:
            raise ValidationError(
                f"failed to Validate {value} as InputCombination", UIMapping
            ) from exception

        if new_combi == self.input_combination:
            return

        # raises a keyError if the combination or a permutation is already mapped
        self._combination_changed(new_combi, self.input_combination)
        super().__setattr__("input_combination", new_combi)

    def __str__(self):
        return str(
            self.dict(
                exclude_defaults=True, include={"input_combination", "target_uinput"}
            )
        )

    if needs_workaround:
        # https://github.com/samuelcolvin/pydantic/issues/1383
        def copy(self: MappingModel, *args, **kwargs) -> MappingModel:
            kwargs["deep"] = True
            copy = super().copy(*args, **kwargs)
            object.__setattr__(copy, "_combination_changed", self._combination_changed)
            return copy

    def format_name(self) -> str:
        """Get the custom-name or a readable representation of the combination."""
        if self.name:
            return self.name

        if (
            self.input_combination == InputCombination.empty_combination()
            or self.input_combination is None
        ):
            return EMPTY_MAPPING_NAME

        return self.input_combination.beautify()

    def has_input_defined(self) -> bool:
        """Whether this mapping defines an event-input."""
        return self.input_combination != InputCombination.empty_combination()

    def is_axis_mapping(self) -> bool:
        """Whether this mapping specifies an output axis."""
        return self.output_type in [EV_ABS, EV_REL]

    def is_wheel_output(self) -> bool:
        """Check if this maps to wheel output."""
        return self.output_code in (
            REL_WHEEL,
            REL_HWHEEL,
        )

    def is_high_res_wheel_output(self) -> bool:
        """Check if this maps to high-res wheel output."""
        return self.output_code in (
            REL_WHEEL_HI_RES,
            REL_HWHEEL_HI_RES,
        )

    def is_analog_output(self):
        return self.mapping_type == MappingType.ANALOG

    def set_combination_changed_callback(self, callback: CombinationChangedCallback):
        self._combination_changed = callback

    def remove_combination_changed_callback(self):
        self._combination_changed = None

    def get_output_type_code(self) -> Optional[Tuple[int, int]]:
        """Returns the output_type and output_code if set,
        otherwise looks the output_symbol up in the keyboard_layout
        return None for unknown symbols and macros
        """
        if self.output_code is not None and self.output_type is not None:
            return self.output_type, self.output_code

        if self.output_symbol and not Parser.is_this_a_macro(self.output_symbol):
            return EV_KEY, keyboard_layout.get(self.output_symbol)

        return None

    def get_output_name_constant(self) -> str:
        """Get the evdev name costant for the output."""
        return get_evdev_constant_name(self.output_type, self.output_code)

    def is_valid(self) -> bool:
        """If the mapping is valid."""
        return not self.get_error()

    def get_error(self) -> Optional[ValidationError]:
        """The validation error or None."""
        try:
            Mapping(**self.dict())
        except ValidationError as exception:
            return exception
        return None

    def get_bus_message(self) -> MappingData:
        """Return an immutable copy for use in the message broker."""
        return MappingData(**self.dict())

    @root_validator
    def validate_mapping_type(cls, values):
        """Overrides the mapping type if the output mapping type is obvious."""
        output_type = values.get("output_type")
        output_code = values.get("output_code")
        output_symbol = values.get("output_symbol")

        if output_type is not None and output_symbol is not None:
            # This is currently only possible when someone edits the preset file by
            # hand. A key-output mapping without an output_symbol, but type and code
            # instead, is valid as well.
            logger.debug("Both output_type and output_symbol are set")

        if output_type != EV_KEY and output_code is not None and not output_symbol:
            values["mapping_type"] = MappingType.ANALOG.value

        if output_type is None and output_code is None and output_symbol:
            values["mapping_type"] = MappingType.KEY_MACRO.value

        if output_type == EV_KEY:
            values["mapping_type"] = MappingType.KEY_MACRO.value

        return values

    Config = Cfg


class Mapping(UIMapping):
    """Holds all the data for mapping an input action to an output action.

    This implements the missing validations from UIMapping.
    """

    # Override Required attributes to enforce they are set
    input_combination: InputCombination
    target_uinput: KnownUinput

    @classmethod
    def from_combination(
        cls,
        input_combination=None,
        target_uinput="keyboard",
        output_symbol="a",
    ):
        """Convenient function to get a valid mapping."""
        if not input_combination:
            input_combination = [{"type": 99, "code": 99, "analog_threshold": 99}]

        return cls(
            input_combination=input_combination,
            target_uinput=target_uinput,
            output_symbol=output_symbol,
        )

    def is_valid(self) -> bool:
        """If the mapping is valid."""
        return True

    @root_validator(pre=True)
    def validate_symbol(cls, values):
        """Parse a macro to check for syntax errors."""
        symbol = values.get("output_symbol")

        if symbol == "":
            values["output_symbol"] = None
            return values

        if symbol is None:
            return values

        symbol = symbol.strip()
        values["output_symbol"] = symbol

        if symbol == DISABLE_NAME:
            return values

        if Parser.is_this_a_macro(symbol):
            mapping_mock = namedtuple("Mapping", values.keys())(**values)
            # raises MacroError
            Parser.parse(symbol, mapping=mapping_mock, verbose=False)
            return values

        code = keyboard_layout.get(symbol)
        if code is None:
            raise OutputSymbolUnknownError(symbol)

        target = values.get("target_uinput")
        if target is not None and not GlobalUInputs.can_default_uinput_emit(
            target, EV_KEY, code
        ):
            raise SymbolNotAvailableInTargetError(symbol, target)

        return values

    @validator("input_combination")
    def only_one_analog_input(cls, combination) -> InputCombination:
        """Check that the input_combination specifies a maximum of one
        analog to analog mapping
        """
        analog_events = [event for event in combination if event.defines_analog_input]
        if len(analog_events) > 1:
            raise OnlyOneAnalogInputError(analog_events)

        return combination

    @validator("input_combination")
    def trigger_point_in_range(cls, combination: InputCombination) -> InputCombination:
        """Check if the trigger point for mapping analog axis to buttons is valid."""
        for input_config in combination:
            if (
                input_config.type == EV_ABS
                and input_config.analog_threshold
                and abs(input_config.analog_threshold) >= 100
            ):
                raise TriggerPointInRangeError(input_config)
        return combination

    @root_validator
    def validate_output_symbol_variant(cls, values):
        """Validate that either type and code or symbol are set for key output."""
        o_symbol = values.get("output_symbol")
        o_type = values.get("output_type")
        o_code = values.get("output_code")
        if o_symbol is None and (o_type is None or o_code is None):
            raise OutputSymbolVariantError()
        return values

    @root_validator
    def validate_output_integrity(cls, values):
        """Validate the output key configuration."""
        symbol = values.get("output_symbol")
        type_ = values.get("output_type")
        code = values.get("output_code")
        if symbol is None:
            # If symbol is "", then validate_symbol changes it to None
            # type and code can be anything
            return values

        if type_ is None and code is None:
            # we have a symbol: no type and code is fine
            return values

        if Parser.is_this_a_macro(symbol):
            # disallow output type and code for macros
            if type_ is not None or code is not None:
                raise MacroButTypeOrCodeSetError()

        if code is not None and code != keyboard_layout.get(symbol) or type_ != EV_KEY:
            raise SymbolAndCodeMismatchError(symbol, code)
        return values

    @root_validator
    def output_matches_input(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Validate that an output type is an axis if we have an input axis.
        And vice versa."""
        assert isinstance(values.get("input_combination"), InputCombination)
        combination: InputCombination = values["input_combination"]

        analog_input_config = combination.find_analog_input_config()
        defines_analog_input = analog_input_config is not None
        output_type = values.get("output_type")
        output_code = values.get("output_code")
        mapping_type = values.get("mapping_type")
        output_symbol = values.get("output_symbol")
        output_key_set = output_symbol or (output_type == EV_KEY and output_code)

        if mapping_type is None:
            # Empty mapping most likely
            return values

        if not defines_analog_input and mapping_type != MappingType.KEY_MACRO.value:
            raise WrongMappingTypeForKeyError()

        if not defines_analog_input and not output_key_set:
            raise MissingMacroOrKeyError()

        if (
            defines_analog_input
            and output_type not in (EV_ABS, EV_REL)
            and output_symbol != DISABLE_NAME
        ):
            raise MissingOutputAxisError(analog_input_config, output_type)

        return values


class MappingData(UIMapping):
    """Like UIMapping, but can be sent over the message broker."""

    Config = ImmutableCfg
    message_type = MessageType.mapping  # allow this to be sent over the MessageBroker

    def __str__(self):
        return str(self.dict(exclude_defaults=True))

    def dict(self, *args, **kwargs):
        """Will not include the message_type."""
        dict_ = super().dict(*args, **kwargs)
        if "message_type" in dict_:
            del dict_["message_type"]
        return dict_



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

# TODO: convert everything to use pathlib.Path

"""Path constants to be used."""


import os
import shutil
from typing import List, Union, Optional

from inputremapper.logging.logger import logger
from inputremapper.user import UserUtils


# TODO maybe this could become, idk, ConfigService and PresetService
class PathUtils:
    @staticmethod
    def config_path() -> str:
        # TODO when proper DI is being done, construct PathUtils and configure it
        #  in the constructor. Then there is no need to recompute the config_path
        #  each time. Tests might have overwritten UserUtils.home.
        xdg_config_home = os.getenv(
            "XDG_CONFIG_HOME", os.path.join(UserUtils.home, ".config")
        )
        return os.path.join(xdg_config_home, "input-remapper-2")

    @staticmethod
    def chown(path):
        """Set the owner of a path to the user."""
        try:
            logger.debug('Chown "%s", "%s"', path, UserUtils.user)
            shutil.chown(path, user=UserUtils.user, group=UserUtils.user)
        except LookupError:
            # the users group was unknown in one case for whatever reason
            shutil.chown(path, user=UserUtils.user)

    @staticmethod
    def touch(path: Union[str, os.PathLike], log=True):
        """Create an empty file and all its parent dirs, give it to the user."""
        if str(path).endswith("/"):
            raise ValueError(f"Expected path to not end with a slash: {path}")

        if os.path.exists(path):
            return

        if log:
            logger.info('Creating file "%s"', path)

        PathUtils.mkdir(os.path.dirname(path), log=False)

        os.mknod(path)
        PathUtils.chown(path)

    @staticmethod
    def mkdir(path, log=True):
        """Create a folder, give it to the user."""
        if path == "" or path is None:
            return

        if os.path.exists(path):
            return

        if log:
            logger.info('Creating dir "%s"', path)

        # give all newly created folders to the user.
        # e.g. if .config/input-remapper/mouse/ is created the latter two
        base = os.path.split(path)[0]
        PathUtils.mkdir(base, log=False)

        os.makedirs(path)
        PathUtils.chown(path)

    @staticmethod
    def split_all(path: Union[os.PathLike, str]) -> List[str]:
        """Split the path into its segments."""
        parts = []
        while True:
            path, tail = os.path.split(path)
            parts.append(tail)
            if path == os.path.sep:
                # we arrived at the root '/'
                parts.append(path)
                break
            if not path:
                # arrived at start of relative path
                break

        parts.reverse()
        return parts

    @staticmethod
    def remove(path):
        """Remove whatever is at the path."""
        if not os.path.exists(path):
            return

        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)

    @staticmethod
    def sanitize_path_component(group_name: str) -> str:
        """Replace characters listed in
        https://en.wikipedia.org/wiki/Filename#Reserved_characters_and_words
        with an underscore.
        """
        for character in '/\\?%*:|"<>':
            if character in group_name:
                group_name = group_name.replace(character, "_")
        return group_name

    @staticmethod
    def get_preset_path(group_name: Optional[str] = None, preset: Optional[str] = None):
        """Get a path to the stored preset, or to store a preset to."""
        presets_base = os.path.join(PathUtils.config_path(), "presets")

        if group_name is None:
            return presets_base

        group_name = PathUtils.sanitize_path_component(group_name)

        if preset is not None:
            # the extension of the preset should not be shown in the ui.
            # if a .json extension arrives this place, it has not been
            # stripped away properly prior to this.
            if not preset.endswith(".json"):
                preset = f"{preset}.json"

        if preset is None:
            return os.path.join(presets_base, group_name)

        return os.path.join(presets_base, group_name, preset)

    @staticmethod
    def get_config_path(*paths) -> str:
        """Get a path in ~/.config/input-remapper/."""
        return os.path.join(PathUtils.config_path(), *paths)



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict

import evdev

from inputremapper.configs.input_config import InputCombination
from inputremapper.injection.mapping_handlers.mapping_handler import (
    MappingHandler,
    HandlerEnums,
)
from inputremapper.input_event import InputEvent


class NullHandler(MappingHandler):
    """Handler which consumes the event and does nothing."""

    def __str__(self):
        return f"NullHandler for {self.mapping.input_combination}<{id(self)}>"

    @property
    def child(self):
        return "Voids all events"

    def needs_wrapping(self) -> bool:
        return False in [
            input_.defines_analog_input for input_ in self.mapping.input_combination
        ]

    def wrap_with(self) -> Dict[InputCombination, HandlerEnums]:
        if not self.mapping.input_combination.defines_analog_input:
            return {self.mapping.input_combination: HandlerEnums.combination}

        assert len(self.mapping.input_combination) > 1, "nees_wrapping ensures this!"
        return {self.mapping.input_combination: HandlerEnums.axisswitch}

    def notify(
        self,
        event: InputEvent,
        source: evdev.InputDevice,
        suppress: bool = False,
    ) -> bool:
        return True

    def reset(self) -> None:
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import os
import unittest
from unittest.mock import MagicMock, patch

import evdev
from evdev.ecodes import (
    EV_KEY,
    EV_ABS,
    BTN_A,
    ABS_X,
    ABS_Y,
    ABS_RX,
    ABS_RY,
    EV_REL,
    REL_X,
    REL_Y,
    REL_HWHEEL_HI_RES,
    REL_WHEEL_HI_RES,
)

from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.configs.keyboard_layout import keyboard_layout
from inputremapper.configs.mapping import Mapping
from inputremapper.configs.preset import Preset
from inputremapper.injection.context import Context
from inputremapper.injection.event_reader import EventReader
from inputremapper.injection.global_uinputs import GlobalUInputs, UInput
from inputremapper.injection.mapping_handlers.mapping_parser import MappingParser
from inputremapper.input_event import InputEvent
from inputremapper.utils import get_device_hash
from tests.lib.fixtures import fixtures
from tests.lib.test_setup import test_setup


@test_setup
class TestEventReader(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.gamepad_source = evdev.InputDevice(fixtures.gamepad.path)
        self.source_hash = fixtures.gamepad.get_device_hash()
        self.global_uinputs = GlobalUInputs(UInput)
        self.mapping_parser = MappingParser(self.global_uinputs)
        self.stop_event = asyncio.Event()
        self.preset = Preset()
        self.forward_uinput = evdev.UInput(name="test-forward-uinput")

        self.global_uinputs.is_service = True
        self.global_uinputs.prepare_all()

    async def setup(self, mapping):
        """Set a EventReader up for the test and run it in the background."""
        context = Context(
            mapping,
            {},
            {self.source_hash: self.forward_uinput},
            self.mapping_parser,
        )
        context.uinput = evdev.UInput()
        event_reader = EventReader(
            context,
            self.gamepad_source,
            self.stop_event,
        )
        asyncio.ensure_future(event_reader.run())
        await asyncio.sleep(0.1)
        return context, event_reader

    async def test_if_single_joystick_then(self):
        # TODO: Move this somewhere more sensible
        # Integration test style for if_single.
        # won't care about the event, because the purpose is not set to BUTTON
        code_a = keyboard_layout.get("a")
        code_shift = keyboard_layout.get("KEY_LEFTSHIFT")
        trigger = evdev.ecodes.BTN_A

        self.preset.add(
            Mapping.from_combination(
                InputCombination(
                    [
                        InputConfig(
                            type=EV_KEY,
                            code=trigger,
                            origin_hash=fixtures.gamepad.get_device_hash(),
                        )
                    ]
                ),
                "keyboard",
                "if_single(key(a), key(KEY_LEFTSHIFT))",
            )
        )
        self.preset.add(
            Mapping.from_combination(
                InputCombination(
                    [
                        InputConfig(
                            type=EV_ABS,
                            code=ABS_Y,
                            analog_threshold=1,
                            origin_hash=fixtures.gamepad.get_device_hash(),
                        )
                    ]
                ),
                "keyboard",
                "b",
            ),
        )

        # left x to mouse x
        config = {
            "input_combination": [
                InputConfig(
                    type=EV_ABS,
                    code=ABS_X,
                    origin_hash=fixtures.gamepad.get_device_hash(),
                )
            ],
            "target_uinput": "mouse",
            "output_type": EV_REL,
            "output_code": REL_X,
        }
        self.preset.add(Mapping(**config))

        # left y to mouse y
        config["input_combination"] = [
            InputConfig(
                type=EV_ABS,
                code=ABS_Y,
                origin_hash=fixtures.gamepad.get_device_hash(),
            )
        ]
        config["output_code"] = REL_Y
        self.preset.add(Mapping(**config))

        # right x to wheel x
        config["input_combination"] = [
            InputConfig(
                type=EV_ABS,
                code=ABS_RX,
                origin_hash=fixtures.gamepad.get_device_hash(),
            )
        ]
        config["output_code"] = REL_HWHEEL_HI_RES
        self.preset.add(Mapping(**config))

        # right y to wheel y
        config["input_combination"] = [
            InputConfig(
                type=EV_ABS,
                code=ABS_RY,
                origin_hash=fixtures.gamepad.get_device_hash(),
            )
        ]
        config["output_code"] = REL_WHEEL_HI_RES
        self.preset.add(Mapping(**config))

        context, _ = await self.setup(self.preset)

        gamepad_hash = get_device_hash(self.gamepad_source)
        self.gamepad_source.push_events(
            [
                InputEvent.key(evdev.ecodes.BTN_Y, 0, gamepad_hash),  # start the macro
                InputEvent.key(trigger, 1, gamepad_hash),  # start the macro
                InputEvent.abs(ABS_Y, 10, gamepad_hash),  # ignored
                InputEvent.key(evdev.ecodes.BTN_B, 2, gamepad_hash),  # ignored
                InputEvent.key(evdev.ecodes.BTN_B, 0, gamepad_hash),  # ignored
                # release the trigger, which runs `then` of if_single
                InputEvent.key(trigger, 0, gamepad_hash),
            ]
        )

        await asyncio.sleep(0.1)
        self.stop_event.set()  # stop the reader

        history = self.global_uinputs.get_uinput("keyboard").write_history
        self.assertIn((EV_KEY, code_a, 1), history)
        self.assertIn((EV_KEY, code_a, 0), history)
        self.assertNotIn((EV_KEY, code_shift, 1), history)
        self.assertNotIn((EV_KEY, code_shift, 0), history)

        # after if_single takes an action, the listener should have been removed
        self.assertSetEqual(context.listeners, set())

    async def test_if_single_joystick_under_threshold(self):
        """Triggers then because the joystick events value is too low."""
        # TODO: Move this somewhere more sensible
        code_a = keyboard_layout.get("a")
        trigger = evdev.ecodes.BTN_A
        self.preset.add(
            Mapping.from_combination(
                InputCombination(
                    [
                        InputConfig(
                            type=EV_KEY,
                            code=trigger,
                            origin_hash=fixtures.gamepad.get_device_hash(),
                        )
                    ]
                ),
                "keyboard",
                "if_single(k(a), k(KEY_LEFTSHIFT))",
            )
        )
        self.preset.add(
            Mapping.from_combination(
                InputCombination(
                    [
                        InputConfig(
                            type=EV_ABS,
                            code=ABS_Y,
                            analog_threshold=1,
                            origin_hash=fixtures.gamepad.get_device_hash(),
                        )
                    ]
                ),
                "keyboard",
                "b",
            ),
        )

        # self.preset.set("gamepad.joystick.left_purpose", BUTTONS)
        # self.preset.set("gamepad.joystick.right_purpose", BUTTONS)
        context, _ = await self.setup(self.preset)

        self.gamepad_source.push_events(
            [
                InputEvent.key(trigger, 1),  # start the macro
                InputEvent.abs(ABS_Y, 1),  # ignored because value too low
                InputEvent.key(trigger, 0),  # stop, only way to trigger `then`
            ]
        )
        await asyncio.sleep(0.1)
        self.assertEqual(len(context.listeners), 0)
        history = self.global_uinputs.get_uinput("keyboard").write_history

        # the key that triggered if_single should be injected after
        # if_single had a chance to inject keys (if the macro is fast enough),
        # so that if_single can inject a modifier to e.g. capitalize the
        # triggering key. This is important for the space cadet shift
        self.assertListEqual(
            history,
            [
                (EV_KEY, code_a, 1),
                (EV_KEY, code_a, 0),
            ],
        )

    @patch.object(Context, "reset")
    async def test_reset_handlers_on_stop(self, reset_mock: MagicMock) -> None:
        await self.setup(self.preset)

        self.stop_event.set()
        await asyncio.sleep(0.1)

        reset_mock.assert_called_once()

    @patch.object(Context, "reset")
    @patch.object(os, "stat")
    async def test_reset_handlers_after_unplugging(
        self,
        stat_mock: MagicMock,
        reset_mock: MagicMock,
    ) -> None:
        await self.setup(self.preset)

        self.gamepad_source.push_events([InputEvent.key(BTN_A, 1)])
        await asyncio.sleep(0.1)
        reset_mock.assert_not_called()

        # unplug the device
        stat_mock().st_nlink = 0

        # It seems that once a device is unplugged, asyncio stops waiting for new input
        # from _source.fileno() or something. I didn't manage to replicate this
        # behavior in tests using the fd of pending_events[fixtures.gamepad][0].fileno()
        # and/or pending_events[fixtures.gamepad][1].fileno(). So instead, I push
        # another event, so that the EventReader makes another iteration.
        self.gamepad_source.push_events([InputEvent.key(BTN_A, 1)])
        await asyncio.sleep(0.1)

        reset_mock.assert_called_once()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import os
import time
import unittest
from unittest.mock import patch, MagicMock

import gi

from inputremapper.injection.global_uinputs import GlobalUInputs, UInput
from inputremapper.injection.mapping_handlers.mapping_parser import MappingParser

gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")
gi.require_version("GtkSource", "4")
gi.require_version("GLib", "2.0")

from inputremapper.configs.global_config import GlobalConfig
from inputremapper.gui.utils import gtk_iteration
from inputremapper.daemon import Daemon

from tests.lib.test_setup import test_setup
from tests.system.gui.gui_test_base import (
    launch,
    start_reader_service,
    clean_up_gui_test,
)


@test_setup
class TestGroupsFromReaderService(unittest.TestCase):
    def patch_os_system(self):
        def os_system(cmd, original_os_system=os.system):
            # instead of running pkexec, fork instead. This will make
            # the reader-service aware of all the test patches
            if "pkexec input-remapper-control --command start-reader-service" in cmd:
                # don't start the reader-service just log that it was.
                self.reader_service_started()
                return 0

            return original_os_system(cmd)

        self.os_system_patch = patch.object(
            os,
            "system",
            os_system,
        )

        # this is already part of the test. we need a bit of patching and hacking
        # because we want to discover the groups as early a possible, to reduce startup
        # time for the application
        self.os_system_patch.start()

    def bootstrap_daemon(self):
        # The daemon gets fresh instances of everything, because as far as I remember
        # it runs in a separate process.
        global_config = GlobalConfig()
        global_uinputs = GlobalUInputs(UInput)
        mapping_parser = MappingParser(global_uinputs)

        return Daemon(
            global_config,
            global_uinputs,
            mapping_parser,
        )

    def patch_daemon(self):
        # don't try to connect, return an object instance of it instead
        self.daemon_connect_patch = patch.object(
            Daemon,
            "connect",
            lambda: self.bootstrap_daemon(),
        )
        self.daemon_connect_patch.start()

    def setUp(self):
        self.reader_service_started = MagicMock()
        self.patch_os_system()
        self.patch_daemon()

        (
            self.user_interface,
            self.controller,
            self.data_manager,
            self.message_broker,
            self.daemon,
            self.global_config,
        ) = launch()

    def tearDown(self):
        clean_up_gui_test(self)
        self.os_system_patch.stop()
        self.daemon_connect_patch.stop()

    def test_knows_devices(self):
        # verify that it is working as expected. The gui doesn't have knowledge
        # of groups until the root-reader-service provides them
        self.data_manager._reader_client.groups.set_groups([])
        gtk_iteration()
        self.reader_service_started.assert_called()
        self.assertEqual(len(self.data_manager.get_group_keys()), 0)

        # start the reader-service delayed
        start_reader_service()
        # perform some iterations so that the reader ends up reading from the pipes
        # which will make it receive devices.
        for _ in range(10):
            time.sleep(0.02)
            gtk_iteration()

        self.assertIn("Foo Device 2", self.data_manager.get_group_keys())
        self.assertIn("Foo Device 2", self.data_manager.get_group_keys())
        self.assertIn("Bar Device", self.data_manager.get_group_keys())
        self.assertIn("gamepad", self.data_manager.get_group_keys())
        self.assertEqual(self.data_manager.active_group.name, "Foo Device")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


import os
import unittest
from unittest import mock

from inputremapper.user import UserUtils
from tests.lib.test_setup import test_setup


def _raise(error):
    raise error


@test_setup
class TestUser(unittest.TestCase):
    def test_get_user(self):
        with mock.patch("os.getlogin", lambda: "foo"):
            self.assertEqual(UserUtils.get_user(), "foo")

        with mock.patch("os.getlogin", lambda: "root"):
            self.assertEqual(UserUtils.get_user(), "root")

        property_mock = mock.Mock()
        property_mock.configure_mock(pw_name="quix")
        with (
            mock.patch("os.getlogin", lambda: _raise(OSError())),
            mock.patch("pwd.getpwuid", return_value=property_mock),
        ):
            os.environ["USER"] = "root"
            os.environ["SUDO_USER"] = "qux"
            self.assertEqual(UserUtils.get_user(), "qux")

            os.environ["USER"] = "root"
            del os.environ["SUDO_USER"]
            os.environ["PKEXEC_UID"] = "1000"
            self.assertNotEqual(UserUtils.get_user(), "root")

    def test_get_home(self):
        property_mock = mock.Mock()
        property_mock.configure_mock(pw_dir="/custom/home/foo")
        with mock.patch("pwd.getpwnam", return_value=property_mock):
            self.assertEqual(UserUtils.get_home("foo"), "/custom/home/foo")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import asyncio
from collections import deque
from typing import Deque

from evdev.ecodes import EV_KEY

from inputremapper.configs.keyboard_layout import keyboard_layout
from inputremapper.injection.macros.argument import ArgumentConfig
from inputremapper.injection.macros.task import Task
from inputremapper.input_event import InputEvent
from inputremapper.logging.logger import logger


class ModTapTask(Task):
    """If pressed long enough in combination with other keys, it turns into a modifier.

    Can be used to make home-row-modifiers.

    Works similar to the default of
    https://github.com/qmk/qmk_firmware/blob/78a0adfbb4d2c4e12f93f2a62ded0020d406243e/docs/tap_hold.md#comparison-comparison
    """

    argument_configs = [
        ArgumentConfig(
            name="default",
            position=0,
            types=[str],
            is_symbol=True,
        ),
        ArgumentConfig(
            name="modifier",
            position=1,
            types=[str],
            is_symbol=True,
        ),
        ArgumentConfig(
            name="tapping_term",
            position=2,
            types=[int, float],
            default=200,
        ),
    ]

    async def run(self, callback) -> None:
        tapping_term = self.get_argument("tapping_term").get_value() / 1000
        jamming_asyncio_events: Deque[asyncio.Event] = deque()

        async def listener(event: InputEvent) -> None:
            trigger = self.mapping.input_combination[-1]
            if event.type_and_code == trigger.type_and_code:
                # We don't block the event that would set _trigger_release_event.
                return

            if event.type != EV_KEY:
                return

            asyncio_event = asyncio.Event()
            jamming_asyncio_events.append(asyncio_event)
            # Make the EventReader wait until the mod_tap macro allows it to continue
            # processing the event. Because we want to wait until mod_tap injected the
            # modifier.
            await asyncio_event.wait()

        self.add_event_listener(listener)

        timeout = asyncio.Task(asyncio.sleep(tapping_term))
        await asyncio.wait(
            [asyncio.Task(self._trigger_release_event.wait()), timeout],
            return_when=asyncio.FIRST_COMPLETED,
        )
        has_timed_out = timeout.done()

        if has_timed_out:
            # The timeout happened before the trigger got released.
            # We therefore modify stuff.
            symbol = self.get_argument("modifier").get_value()
            logger.debug("Modifying with %s", symbol)
        else:
            # The trigger got released before the timeout.
            # We therefore do not modify stuff.
            symbol = self.get_argument("default").get_value()
            logger.debug("Writing default %s", symbol)

        code = keyboard_layout.get(symbol)
        callback(EV_KEY, code, 1)
        await self.keycode_pause()

        # Now that we know if the key was pressed with the intention of modifying other
        # keys, we can let the jammed keys go on their journey through the handlers.
        # Those other handlers may map them to other keys and stuff.
        while len(jamming_asyncio_events) > 0:
            asyncio_event = jamming_asyncio_events.popleft()
            asyncio_event.set()
            await self.keycode_pause()
            await self.throttle()
            # While we are emptying the queue, more events might still arrive and add
            # to the queue.

        # We remove this as late as possible, because if more keys are pressed while
        # jamming_asyncio_events is still being taken care of, they should wait until
        # all is done. This ensures the order of all events that are pressed, until
        # mod_tap is completely finished.
        self.remove_event_listener(listener)

        # Keep the modifier pressed until the input/trigger is released
        await self._trigger_release_event.wait()
        callback(EV_KEY, code, 0)

        await self.keycode_pause()

    async def throttle(self) -> None:
        # In case the keycode_pause ist set to 0ms, we need to give the event handlers
        # a chance to inject the withheld events, before we go on. This ensures the
        # correct order of injections. Since we are using asyncio, something like
        # `callback(EV_KEY, code, 0)` might be faster than the event handlers, even if
        # it is the last step of the macro.
        if self.mapping.macro_key_sleep_ms == 0:
            await asyncio.sleep(0.01)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


def check_dependencies() -> None:
    print("Checking dependencies")
    try:
        import gi

        gi.require_version("Gdk", "3.0")
        gi.require_version("GLib", "2.0")
        gi.require_version("Gst", "1.0")
        gi.require_version("Gtk", "3.0")
        gi.require_version("GtkSource", "4")
        from gi.repository import GObject, Gtk, Gst, Gdk, GLib, Pango, Gio, GtkSource
        import evdev
        import psutil
        import dasbus
        import pygobject
        import pydantic

        print("All required Python modules found")
    except ImportError as e:
        print(f"\033[93mMissing Python module: {e}\033[0m")
    except Exception as e:
        print(f"\033[93mException while checking dependencies: {e}\033[0m")



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


"""Starts injecting keycodes based on the configuration.

https://github.com/dasbus-project/dasbus/tree/master/examples/03_helloworld
"""

import atexit
import json
import os
import sys
import time
from pathlib import PurePath
from typing import Dict, Optional, Protocol

import gi
from dasbus.error import DBusError
from dasbus.connection import SystemMessageBus
from dasbus.identifier import DBusServiceIdentifier
from dasbus.loop import EventLoop

from inputremapper.configs.global_config import GlobalConfig
from inputremapper.configs.keyboard_layout import keyboard_layout
from inputremapper.configs.paths import PathUtils
from inputremapper.configs.preset import Preset
from inputremapper.groups import groups
from inputremapper.injection.global_uinputs import GlobalUInputs
from inputremapper.injection.injector import Injector, InjectorState
from inputremapper.injection.macros.macro import macro_variables
from inputremapper.injection.mapping_handlers.mapping_parser import MappingParser
from inputremapper.logging.logger import logger
from inputremapper.user import UserUtils

gi.require_version("GLib", "2.0")


SYSTEM_BUS = SystemMessageBus()

DAEMON = DBusServiceIdentifier(
    namespace=("inputremapper", "Control"),
    message_bus=SYSTEM_BUS,
)

# timeout in milliseconds
BUS_TIMEOUT = 10000


class AutoloadHistory:
    """Contains the autoloading history and constraints."""

    def __init__(self):
        """Construct this with an empty history."""
        # preset of device -> (timestamp, preset)
        self._autoload_history = {}

    def remember(self, group_key: str, preset: str):
        """Remember when this preset was autoloaded for the device."""
        self._autoload_history[group_key] = (time.time(), preset)

    def forget(self, group_key: str):
        """The injection was stopped or started by hand."""
        if group_key in self._autoload_history:
            del self._autoload_history[group_key]

    def may_autoload(self, group_key: str, preset: str):
        """Check if this autoload would be redundant.

        This is needed because udev triggers multiple times per hardware
        device, and because it should be possible to stop the injection
        by unplugging the device if the preset goes wrong or if input-remapper
        has some bug that prevents the computer from being controlled.

        For that unplug and reconnect the device twice within a 15 seconds
        timeframe which will then not ask for autoloading again. Wait 3
        seconds between replugging.
        """
        if group_key not in self._autoload_history:
            return True

        if self._autoload_history[group_key][1] != preset:
            return True

        # bluetooth devices go to standby mode after some time. After a
        # certain time of being disconnected it should be legit to autoload
        # again. It takes 2.5 seconds for me when quickly replugging my usb
        # mouse until the daemon is asked to autoload again. Redundant calls
        # by udev to autoload for the device seem to happen within 0.2
        # seconds in my case.
        now = time.time()
        threshold = 15  # seconds
        if self._autoload_history[group_key][0] < now - threshold:
            return True

        return False


class DaemonProxy(Protocol):  # pragma: no cover
    """The interface provided over the dbus."""

    def stop_injecting(self, group_key: str) -> None: ...

    def get_state(self, group_key: str) -> InjectorState: ...

    def start_injecting(self, group_key: str, preset: str) -> bool: ...

    def stop_all(self) -> None: ...

    def set_config_dir(self, config_dir: str) -> None: ...

    def autoload(self) -> None: ...

    def autoload_single(self, group_key: str) -> None: ...

    def hello(self, out: str) -> str: ...

    def quit(self) -> None: ...


class Daemon:
    """Starts injecting keycodes based on the configuration.

    Can be talked to either over dbus or by instantiating it.

    The Daemon may not have any knowledge about the logged in user, so it
    can't read any config files. It has to be told what to do and will
    continue to do so afterwards, but it can't decide to start injecting
    on its own.
    """

    # https://dbus.freedesktop.org/doc/dbus-specification.html#type-system
    __dbus_xml__ = f"""
        <node>
            <interface name='{DAEMON.interface_name}'>
                <method name='stop_injecting'>
                    <arg type='s' name='group_key' direction='in'/>
                </method>
                <method name='get_state'>
                    <arg type='s' name='group_key' direction='in'/>
                    <arg type='s' name='response' direction='out'/>
                </method>
                <method name='start_injecting'>
                    <arg type='s' name='group_key' direction='in'/>
                    <arg type='s' name='preset' direction='in'/>
                    <arg type='b' name='response' direction='out'/>
                </method>
                <method name='stop_all'>
                </method>
                <method name='set_config_dir'>
                    <arg type='s' name='config_dir' direction='in'/>
                </method>
                <method name='autoload'>
                </method>
                <method name='autoload_single'>
                    <arg type='s' name='group_key' direction='in'/>
                </method>
                <method name='hello'>
                    <arg type='s' name='out' direction='in'/>
                    <arg type='s' name='response' direction='out'/>
                </method>
                <method name='quit'>
                </method>
            </interface>
        </node>
    """

    def __init__(
        self,
        global_config: GlobalConfig,
        global_uinputs: GlobalUInputs,
        mapping_parser: MappingParser,
    ) -> None:
        """Constructs the daemon."""
        logger.debug("Creating daemon")

        self.global_config = global_config
        self.global_uinputs = global_uinputs
        self.mapping_parser = mapping_parser

        self.injectors: Dict[str, Injector] = {}

        self.config_dir = None

        if UserUtils.user != "root":
            # If started via sudo, UserUtils.user is the actual user, and the config
            # path is valid. If running as a systemd service, it is root, and the path
            # invalid.
            self.set_config_dir(PathUtils.get_config_path())

        # check privileges
        if os.getuid() != 0:
            logger.warning("The service usually needs elevated privileges")

        self.autoload_history = AutoloadHistory()
        self.refreshed_devices_at = 0

        atexit.register(self.stop_all)

        # initialize stuff that is needed alongside the daemon process
        macro_variables.start()

    @classmethod
    def connect(cls, fallback: bool = True) -> Optional[DaemonProxy]:
        """Get a proxy to start and stop injecting keystrokes.

        Parameters
        ----------
        fallback
            If true, starts the daemon via pkexec if it cannot connect.
        """
        try:
            proxy = DAEMON.get_proxy()
            # Calling proxy.Introspect to check if the proxy is running
            # https://dbus.freedesktop.org/doc/dbus-tutorial.html#introspection
            proxy.Introspect(timeout=BUS_TIMEOUT)
            logger.info("Connected to the service")
        except DBusError as error:
            if not fallback:
                logger.error("Service not running? %s", error)
                return None

            logger.info("Starting the service")
            # Blocks until pkexec is done asking for the password.
            # Runs via input-remapper-control so that auth_admin_keep works
            # for all pkexec calls of the gui
            debug = " -d" if logger.is_debug() else ""
            cmd = f"pkexec input-remapper-control --command start-daemon {debug}"

            # using pkexec will also cause the service to continue running in
            # the background after the gui has been closed, which will keep
            # the injections ongoing

            logger.debug("Running `%s`", cmd)
            os.system(cmd)
            time.sleep(0.2)

            # try a few times if the service was just started
            for attempt in range(3):
                try:
                    proxy = DAEMON.get_proxy()
                    proxy.Introspect(timeout=BUS_TIMEOUT)
                    break
                except DBusError as error:
                    logger.debug("Attempt %d to reach the service failed:", attempt + 1)
                    logger.debug('"%s"', error)
                time.sleep(0.2)
            else:
                logger.error("Failed to connect to the service")
                sys.exit(8)

        if UserUtils.user != "root":
            config_path = PathUtils.get_config_path()
            logger.debug('Telling service about "%s"', config_path)
            proxy.set_config_dir(PathUtils.get_config_path(), timeout=2000)

        return proxy

    def publish(self) -> None:
        """Make the dbus interface available."""
        try:
            SYSTEM_BUS.publish_object(DAEMON.object_path, self)
            SYSTEM_BUS.register_service(DAEMON.service_name)
        except ConnectionError as error:
            logger.error("Is the service already running? (%s)", str(error))
            sys.exit(9)

    def run(self) -> None:
        """Start the daemons loop. Blocks until the daemon stops."""
        loop = EventLoop()
        logger.debug("Running daemon")
        loop.run()

    def refresh(self, group_key: Optional[str] = None) -> None:
        """Refresh groups if the specified group is unknown.

        Parameters
        ----------
        group_key
            unique identifier used by the groups object
        """
        now = time.time()
        if now - 10 > self.refreshed_devices_at:
            logger.debug("Refreshing because last info is too old")
            # it may take a bit of time until devices are visible after changes
            time.sleep(0.1)
            groups.refresh()
            self.refreshed_devices_at = now
            return

        if not groups.find(key=group_key):
            logger.debug('Refreshing because "%s" is unknown', group_key)
            time.sleep(0.1)
            groups.refresh()
            self.refreshed_devices_at = now

    def stop_injecting(self, group_key: str) -> None:
        """Stop injecting the preset mappings for a single device."""
        if self.injectors.get(group_key) is None:
            logger.debug(
                'Tried to stop injector, but none is running for group "%s"',
                group_key,
            )
            return

        self.injectors[group_key].stop_injecting()
        self.autoload_history.forget(group_key)

    def get_state(self, group_key: str) -> InjectorState:
        """Get the injectors state."""
        injector = self.injectors.get(group_key)
        return injector.get_state() if injector else InjectorState.UNKNOWN

    def set_config_dir(self, config_dir: str) -> None:
        """All future operations will use this config dir.

        Existing injections (possibly of the previous user) will be kept
        alive, call stop_all to stop them.

        Parameters
        ----------
        config_dir
            This path contains config.json, xmodmap.json and the
            presets directory
        """
        config_path = PurePath(config_dir, "config.json")
        if not os.path.exists(config_path):
            logger.error('"%s" does not exist', config_path)
            return

        self.config_dir = config_dir
        self.global_config.load_config(str(config_path))

    def _autoload(self, group_key: str) -> None:
        """Check if autoloading is a good idea, and if so do it.

        Parameters
        ----------
        group_key
            unique identifier used by the groups object
        """
        self.refresh(group_key)

        group = groups.find(key=group_key)
        if group is None:
            # even after groups.refresh, the device is unknown, so it's
            # either not relevant for input-remapper, or not connected yet
            return

        preset = self.global_config.get_autoload_preset(group.key)

        if preset is None:
            # no autoloading is configured for this device
            return

        if not isinstance(preset, str):
            # maybe another dict or something, who knows. Broken config
            logger.error("Expected a string for autoload, but got %s", preset)
            return

        logger.info('Autoloading for "%s"', group.key)

        if not self.autoload_history.may_autoload(group.key, preset):
            logger.info(
                'Not autoloading the same preset "%s" again for group "%s"',
                preset,
                group.key,
            )
            return

        self.start_injecting(group.key, preset)
        self.autoload_history.remember(group.key, preset)

    def autoload_single(self, group_key: str) -> None:
        """Inject the configured autoload preset for the device.

        If the preset is already being injected, it won't autoload it again.

        Parameters
        ----------
        group_key
            unique identifier used by the groups object
        """
        # avoid some confusing logs and filter obviously invalid requests
        if group_key.startswith("input-remapper"):
            return

        logger.info('Request to autoload for "%s"', group_key)

        if self.config_dir is None:
            logger.error(
                'Request to autoload "%s" before a user told the service about their '
                "session using set_config_dir",
                group_key,
            )
            return

        self._autoload(group_key)

    def autoload(self) -> None:
        """Load all autoloaded presets for the current config_dir.

        If the preset is already being injected, it won't autoload it again.
        """
        if self.config_dir is None:
            logger.error(
                "Request to autoload all before a user told the service about their "
                "session using set_config_dir",
            )
            return

        autoload_presets = list(self.global_config.iterate_autoload_presets())

        logger.info("Autoloading for all devices")

        if len(autoload_presets) == 0:
            logger.error("No presets configured to autoload")
            return

        for group_key, _ in autoload_presets:
            self._autoload(group_key)

    def start_injecting(self, group_key: str, preset_name: str) -> bool:
        """Start injecting the preset for the device.

        Returns True on success. If an injection is already ongoing for
        the specified device it will stop it automatically first.

        Parameters
        ----------
        group_key
            The unique key of the group
        preset_name
            The name of the preset
        """
        logger.info('Request to start injecting for "%s"', group_key)

        self.refresh(group_key)

        if self.config_dir is None:
            logger.error(
                "Request to start an injectoin before a user told the service about "
                "their session using set_config_dir",
            )
            return False

        group = groups.find(key=group_key)

        if group is None:
            logger.error('Could not find group "%s"', group_key)
            return False

        preset_path = PurePath(
            self.config_dir,
            "presets",
            PathUtils.sanitize_path_component(group.name),
            f"{preset_name}.json",
        )

        # Path to a dump of the xkb mappings, to provide more human
        # readable keys in the correct keyboard layout to the service.
        # The service cannot use `xmodmap -pke` because it's running via
        # systemd.
        xmodmap_path = os.path.join(self.config_dir, "xmodmap.json")
        try:
            with open(xmodmap_path, "r") as file:
                # do this for each injection to make sure it is up to
                # date when the system layout changes.
                xmodmap = json.load(file)
                logger.debug('Using keycodes from "%s"', xmodmap_path)

                # this creates the keyboard_layout._xmodmap, which we need to do now
                # otherwise it might be created later which will override the changes
                # we do here.
                # Do we really need to lazyload in the keyboard_layout?
                # this kind of bug is stupid to track down
                keyboard_layout.get_name(0)
                keyboard_layout.update(xmodmap)
                # the service now has process wide knowledge of xmodmap
                # keys of the users session
        except FileNotFoundError:
            logger.error('Could not find "%s"', xmodmap_path)

        preset = Preset(preset_path)

        try:
            preset.load()
        except FileNotFoundError as error:
            logger.error(str(error))
            return False

        for mapping in preset:
            # only create those uinputs that are required to avoid
            # confusing the system. Seems to be especially important with
            # gamepads, because some apps treat the first gamepad they found
            # as the only gamepad they'll ever care about.
            self.global_uinputs.prepare_single(mapping.target_uinput)

        if self.injectors.get(group_key) is not None:
            self.stop_injecting(group_key)

        try:
            injector = Injector(
                group,
                preset,
                self.mapping_parser,
            )
            injector.start()
            self.injectors[group.key] = injector
        except OSError:
            # I think this will never happen, probably leftover from
            # some earlier version
            return False

        return True

    def stop_all(self) -> None:
        """Stop all injections."""
        logger.info("Stopping all injections")
        for group_key in list(self.injectors.keys()):
            self.stop_injecting(group_key)

    def hello(self, out: str) -> str:
        """Used for tests."""
        logger.info('Received "%s" from client', out)
        return out

    def quit(self) -> None:
        """Stop the process."""
        # Beware, that stop_all will also be called via atexit.register(self.stop_all)
        logger.info("Got command to stop the daemon process")
        sys.exit(0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import unittest

from evdev._ecodes import (
    REL_Y,
    EV_REL,
    REL_HWHEEL,
    REL_HWHEEL_HI_RES,
    REL_X,
    REL_WHEEL,
    REL_WHEEL_HI_RES,
    KEY_A,
    EV_KEY,
)

from inputremapper.configs.validation_errors import MacroError
from inputremapper.injection.macros.parse import Parser
from tests.lib.test_setup import test_setup
from tests.unit.test_macros.macro_test_base import DummyMapping, MacroTestBase


@test_setup
class TestMouse(MacroTestBase):
    async def test_mouse_acceleration(self):
        # There is a tiny float-rounding error that can break the test, therefore I use
        # 0.09001 to make it more robust.
        await self._run_macro("mouse(up, 10, 0.09001)", 0.1)
        self.assertEqual(
            [
                (EV_REL, REL_Y, -2),
                (EV_REL, REL_Y, -3),
                (EV_REL, REL_Y, -4),
                (EV_REL, REL_Y, -4),
                (EV_REL, REL_Y, -5),
            ],
            self.result,
        )

    async def test_rate(self):
        # It should move 200 times per second by 1px, for 0.2 seconds.
        rel_rate = 200
        time = 0.2
        speed = 1
        expected_movement = time * rel_rate * speed

        await self._run_macro(f"mouse(down, {speed})", time, rel_rate)
        total_movement = sum(event[2] for event in self.result)
        self.assertAlmostEqual(float(total_movement), expected_movement, delta=1)

    async def test_slow_movement(self):
        await self._run_macro("mouse(down, 0.1)", 0.2, 200)
        total_movement = sum(event[2] for event in self.result)
        self.assertAlmostEqual(total_movement, 4, delta=1)

    async def test_mouse_xy_acceleration_1(self):
        await self._run_macro("mouse_xy(2, -10, 0.09001)", 0.1)
        self.assertEqual(
            [
                (EV_REL, REL_Y, -2),
                (EV_REL, REL_Y, -3),
                (EV_REL, REL_Y, -4),
                (EV_REL, REL_Y, -4),
                (EV_REL, REL_Y, -5),
            ],
            self._get_y_movement(),
        )
        self.assertEqual(
            [
                (EV_REL, REL_X, 1),
                (EV_REL, REL_X, 1),
                (EV_REL, REL_X, 1),
            ],
            self._get_x_movement(),
        )

    async def test_mouse_xy_acceleration_2(self):
        await self._run_macro("mouse_xy(10, -2, 0.09001)", 0.1)
        self.assertEqual(
            [
                (EV_REL, REL_Y, -1),
                (EV_REL, REL_Y, -1),
                (EV_REL, REL_Y, -1),
            ],
            self._get_y_movement(),
        )
        self.assertEqual(
            [
                (EV_REL, REL_X, 2),
                (EV_REL, REL_X, 3),
                (EV_REL, REL_X, 4),
                (EV_REL, REL_X, 4),
                (EV_REL, REL_X, 5),
            ],
            self._get_x_movement(),
        )

    async def test_mouse_xy_only_x(self):
        await self._run_macro("mouse_xy(x=10, acceleration=1)", 0.1)
        self.assertEqual([], self._get_y_movement())
        self.assertEqual(
            [
                (EV_REL, REL_X, 10),
                (EV_REL, REL_X, 10),
                (EV_REL, REL_X, 10),
                (EV_REL, REL_X, 10),
                (EV_REL, REL_X, 10),
                (EV_REL, REL_X, 10),
            ],
            self._get_x_movement(),
        )

    async def test_mouse_xy_only_y(self):
        await self._run_macro("mouse_xy(y=10)", 0.1)
        self.assertEqual([], self._get_x_movement())
        self.assertEqual(
            [
                (EV_REL, REL_Y, 10),
                (EV_REL, REL_Y, 10),
                (EV_REL, REL_Y, 10),
                (EV_REL, REL_Y, 10),
                (EV_REL, REL_Y, 10),
                (EV_REL, REL_Y, 10),
            ],
            self._get_y_movement(),
        )

    async def test_wheel_left(self):
        wheel_speed = 60
        sleep = 0.1
        await self._run_macro(f"wheel(left, {wheel_speed})", sleep)

        self.assertIn((EV_REL, REL_HWHEEL, 1), self.result)
        self.assertIn((EV_REL, REL_HWHEEL_HI_RES, 60), self.result)

        expected_num_hires_events = sleep * DummyMapping.rel_rate
        expected_num_wheel_events = int(expected_num_hires_events / 120 * wheel_speed)
        actual_num_wheel_events = self.result.count((EV_REL, REL_HWHEEL, 1))
        actual_num_hires_events = self.result.count(
            (
                EV_REL,
                REL_HWHEEL_HI_RES,
                wheel_speed,
            )
        )

        self.assertGreater(
            actual_num_wheel_events,
            expected_num_wheel_events * 0.9,
        )
        self.assertLess(
            actual_num_wheel_events,
            expected_num_wheel_events * 1.1,
        )
        self.assertGreater(
            actual_num_hires_events,
            expected_num_hires_events * 0.9,
        )
        self.assertLess(
            actual_num_hires_events,
            expected_num_hires_events * 1.1,
        )

    async def test_wheel_up(self):
        await self._run_macro("wheel(up, 60)", 0.1)
        self.assertIn((EV_REL, REL_WHEEL, 1), self.result)
        self.assertIn((EV_REL, REL_WHEEL_HI_RES, 60), self.result)

    async def test_wheel_down(self):
        await self._run_macro("wheel(down, 60)", 0.1)
        self.assertIn((EV_REL, REL_WHEEL, -1), self.result)
        self.assertIn((EV_REL, REL_WHEEL_HI_RES, -60), self.result)

    async def test_wheel_right(self):
        await self._run_macro("wheel(right, 60)", 0.1)
        self.assertIn((EV_REL, REL_HWHEEL, -1), self.result)
        self.assertIn((EV_REL, REL_HWHEEL_HI_RES, -60), self.result)

    async def test_mouse_releases(self):
        await self._run_macro("mouse(down, 1).key(a)", 0.1)
        self.assertEqual(self.result[-2:], [(EV_KEY, KEY_A, 1), (EV_KEY, KEY_A, 0)])

    async def test_mouse_xy_releases(self):
        await self._run_macro("mouse_xy(1, 1, 1).key(a)", 0.1)
        self.assertEqual(self.result[-2:], [(EV_KEY, KEY_A, 1), (EV_KEY, KEY_A, 0)])

    async def test_wheel_releases(self):
        await self._run_macro("wheel(down, 1).key(a)", 0.1)
        self.assertEqual(self.result[-2:], [(EV_KEY, KEY_A, 1), (EV_KEY, KEY_A, 0)])

    async def test_raises_error(self):
        Parser.parse("mouse(up, 3)", self.context)  # no error
        Parser.parse("mouse(up, speed=$a)", self.context)  # no error
        self.assertRaises(MacroError, Parser.parse, "mouse(3, up)", self.context)
        Parser.parse("wheel(left, 3)", self.context)  # no error
        self.assertRaises(MacroError, Parser.parse, "wheel(3, left)", self.context)

    def _get_x_movement(self):
        return [event for event in self.result if event[1] == REL_X]

    def _get_y_movement(self):
        return [event for event in self.result if event[1] == REL_Y]

    async def _run_macro(
        self,
        code: str,
        time: float,
        rel_rate: int = DummyMapping.rel_rate,
    ):
        dummy_mapping = DummyMapping()
        dummy_mapping.rel_rate = rel_rate
        macro = Parser.parse(
            code,
            self.context,
            dummy_mapping,
        )
        macro.press_trigger()
        asyncio.ensure_future(macro.run(self.handler))

        await asyncio.sleep(time)
        self.assertTrue(macro.tasks[0].is_holding())
        macro.release_trigger()
        await asyncio.sleep(0.05)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


import unittest

from evdev.ecodes import EV_KEY

from inputremapper.configs.keyboard_layout import keyboard_layout
from inputremapper.configs.validation_errors import MacroError
from inputremapper.injection.macros.parse import Parser
from tests.lib.test_setup import test_setup
from tests.unit.test_macros.macro_test_base import MacroTestBase, DummyMapping


@test_setup
class TestModify(MacroTestBase):
    async def test_modify(self):
        code_a = keyboard_layout.get("a")
        code_b = keyboard_layout.get("b")
        code_c = keyboard_layout.get("c")
        macro = Parser.parse(
            "set(foo, b).modify($foo, modify(a, key(c)))",
            self.context,
            DummyMapping,
        )
        await macro.run(self.handler)
        self.assertListEqual(
            self.result,
            [
                (EV_KEY, code_b, 1),
                (EV_KEY, code_a, 1),
                (EV_KEY, code_c, 1),
                (EV_KEY, code_c, 0),
                (EV_KEY, code_a, 0),
                (EV_KEY, code_b, 0),
            ],
        )

    async def test_raises_error(self):
        self.assertRaises(MacroError, Parser.parse, "modify(asdf, k(a))", self.context)


if __name__ == "__main__":
    unittest.main()
